# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2015-2016 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/
"""
configuration used to run freeart reconstruction.
The interpreter will have an instance of :class:`_ReconsConfig` which can either
be a :class:`TxConfig` or a :class:`FluoConfig`

Those classes can saved and read to configuration files.
freeart is using a `silx.io.dictdump.DictDump` object. So those configuration
can be saved into (.ini, .cfg, .h5) files.
To save them you simply have to call :meth:`save` and :meth:`read` to load them.

Transmission configuration
--------------------------

This is an example of a configuration file (.cfg) and what are the main
parameters.

.. code-block:: txt

    [general_settings]
    date  = 14/11/2017
    reconstruction_type = Transmission
    freeart_version = 3.0.0
    precision = simple

    [normalization]
    i0 = 1.0
    i0_index = 
    rotation_center = 327
    computeminuslog = True
    normalizationi0fromafile = False

    [reconstruction_properties]
    include_last_angle = False
    solid_angle_is_off = False
    voxel_size = 1.0
    oversampling = 2
    relaxation_factor = 0.00311526479751
    bean_calculation_method = 0

    [reduction_data]
    definition_reducted_by = 1
    projection_number_reducted_by = 1

    [projection_information]
    projections_sel = 0:321
    acqui_inv = False
    min_angle = 0.0
    max_angle = 6.28318530718

    [data_source_tx]
    sino_file = /users/toto/sinogram.edf
    sino_file_index = 0


Fluorescence configuration
--------------------------

This is an example of a configuration file (.cfg) and what are the main
parameters.


.. code-block:: txt

    [general_settings]
    date  = 14/11/2017
    reconstruction_type = Fluorescence
    freeart_version = 3.0.0
    precision = double

    [normalization]
    i0 = 1.0
    i0_index =
    rotation_center =
    normalizationi0fromafile = False

    [reconstruction_properties]
    include_last_angle = False
    solid_angle_is_off = False
    voxel_size = 1.0
    oversampling = 2
    relaxation_factor = 0.00311526479751
    bean_calculation_method = 0
    e0 = 1.0
    outgoing_bean_calculation_method = 0

    [reduction_data]
    definition_reducted_by = 1
    projection_number_reducted_by = 1

    [projection_information]
    projections_sel = :
    acqui_inv = False
    min_angle = 0.0
    max_angle = 6.28318530718

    [detector_setup]
    det_pos_x = 1000.0
    det_pos_y = 1000.0
    det_pos_z = 0.0
    det_width = 1.0

    [data_source_fluo]
    materials_file =
    absorption_file = absorptionMatrix.edf
    samp_composition_file =
    absorption_file_is_a_sinogram = False
    samp_composition_file_index =
    absorption_file_index = 0
    materials_file_index =

    [fluo_sino_file_0]
    self_absorption_file_0 = absorptionFile.edf
    data_set_index_0 = 0
    data_name_0 = \0
    data_physical_element_0 = H
    ef_0 = 1.0
    file_path = fluoSinogram.edf


.. note:: In order to generate for configuration file you can use tomogui (http://gitlab.esrf.fr/tomoTools/tomogui)

"""

__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "30/10/2017"

from .h5ConfigIO import H5ConfigReader, H5ConfigWriter
from .iniConfigIO import IniConfigReader, IniConfigWriter
import os


INI_FILE_EXT = '.cfg', '.ini'
"""Extension compatible with the dict dump"""

HDF5_FILE_EXT = '.h5', '.hdf5'
"""Valid file extension for hdf5"""


def read(filePath):
    """
    Read the configuration stored in the given file path

    :param filePath: file to read
    :return: the object containing the information for the reconstruction
    :rtype: _ReconsConfig instance
    """
    assert os.path.isfile(filePath)
    fn, file_extension = os.path.splitext(filePath)
    if file_extension in INI_FILE_EXT:
        reader = IniConfigReader()
    elif file_extension in HDF5_FILE_EXT:
        reader = H5ConfigReader()
    else:
        raise IOError('File extension not managed, can\'t read the file')

    return reader.read(filePath)


def save(configuration, filePath, overwrite=False, merge=False):
    """
    save the given configuration into the given filePath

    :param configuration: the configuration to write
    :param filePath: storage location
    """
    if filePath.lower().endswith(INI_FILE_EXT):
        writer = IniConfigWriter()
    elif filePath.lower().endswith(HDF5_FILE_EXT):
        writer = H5ConfigWriter()
    else:
        raise IOError('File extension not managed, can\'t write the file')
    writer.write(filePath=filePath, reconsConfiguration=configuration,
                 overwrite=overwrite, merge=merge)



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/

__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "01/09/2016"


import os
import tempfile

import numpy

from freeart.configuration import config
from freeart.utils import reconstrutils


def writeGeneralSettings(f, reconsType):
    with open(f, 'a') as myFile:
        myFile.write("[general_settings]\n")
        myFile.write("reconstruction_type = " + config._ReconsConfig.FLUO_ID+ "\n")
        myFile.write("freeart_version = 3.0.0\n")
        myFile.write("precision = double\n")


def writeFluoGeneral(f, absFileIsASino, absFile, sampComposition,
                     materialsFile):
    assert type(absFileIsASino) is bool
    assert type(absFile) is str

    with open(f, 'a') as myFile:
        myFile.write("[data_source_fluo]\n")
        myFile.write("absorption_file_is_a_sinogram = " + str(absFileIsASino) + "\n")
        myFile.write("absorption_file = " + absFile + "\n")
        myFile.write("absorption_file_index = 0\n")
        _sam = sampComposition
        if _sam is None:
            _sam = ''
        myFile.write("samp_composition_file = " + _sam + "\n")
        myFile.write("samp_composition_file_index = 0\n")
        _mat = materialsFile
        if _mat is None:
            _mat = ''
        myFile.write("materials_file = " + _mat + "\n")
        myFile.write('materials_file_index = 0\n')
        myFile.write("\n")


def addSinoFile(f, dim, nToAdd, physElmts, selfAbsPath, tempdir=None):
    """
    Simulate a fluorescence sinogram to reconstruct.
    
    :param f: 
    :param nToAdd: 
    :param physelmt: 
    :return:
    """
    _tempdir = tempdir
    if _tempdir is None:
        _tempdir = tempfile.mkdtemp()

    with open(f, 'a') as myFile:
        for i in range(nToAdd):
            myFile.write("[fluo_sino_file_" + str(i) + "]\n")
            sinoPath = os.path.join(tempdir, 'selfAbsMat_' + str(i) + ".edf")
            reconstrutils.saveMatrix(numpy.random.rand(dim, dim),
                                     sinoPath)
            myFile.write("fluo_sino_url_0 =  " + sinoPath + "\n")
            myFile.write("data_set_index_0 = 0\n")
            myFile.write("data_name_0 = " + physElmts[i] + "\n")
            myFile.write("data_physical_element_0 = " + physElmts[i] + "\n")
            myFile.write("ef_0 = 1.0\n")
            myFile.write("self_absorption_file_0 = " + selfAbsPath + "\n")
            myFile.write('\n')

    return tempdir


def addNormalization(f, rotCenter, normI0FrmFile, i0, computeMinusLog):
    with open(f, 'a') as myFile:
        myFile.write("[normalization]\n")
        myFile.write("rotation_center = " + str(rotCenter) + "\n")
        myFile.write("normalizei0fromafile = " + str(normI0FrmFile) + "\n")
        myFile.write("i0 = " + str(i0) + "\n")
        myFile.write("computeminuslog = " + str(computeMinusLog) + "\n")
        myFile.write('\n')


def addReconsProperties(f, voxelSize=1.0, oversampling=2, relaxationFactor=1,
                        beamCalMeth=0, outBeamCalc=0, solidAngOff=False,
                        includeLatAng=False, e0=1):
    with open(f, 'a') as myFile:
        myFile.write("[reconstruction_properties]\n")
        myFile.write("voxel_size = " + str(voxelSize) + "\n")
        myFile.write("oversampling = " + str(oversampling) + "\n")
        myFile.write("relaxation_factor = " + str(relaxationFactor) + "\n")
        myFile.write("bean_calculation_method = " + str(beamCalMeth) + "\n")
        myFile.write("outgoing_bean_calculation_method = " + str(outBeamCalc) + "\n")
        myFile.write("solid_angle_is_off = " + str(solidAngOff) + "\n")
        myFile.write("include_last_angle = " + str(includeLatAng) + "\n")
        myFile.write("e0 = " + str(e0) + "\n")
        myFile.write('\n')


def addReductionData(f, definReduc=1, projReduct=1):
    with open(f, 'a') as myFile:
        myFile.write("[reduction_data]\n")
        myFile.write("definition_reducted_by = " + str(definReduc) + "\n")
        myFile.write("projection_number_reducted_by = " + str(projReduct) + "\n")
        myFile.write('\n')


def addProjInfo(f, startProj, endProj, minAngle=0, maxAngle=numpy.pi*2.0, invertAcqui=False):
    with open(f, 'a') as myFile:
        myFile.write("[projection_information]\n")
        myFile.write("min_angle = " + str(minAngle) + "\n")
        myFile.write("max_angle = " + str(maxAngle) + "\n")
        myFile.write("projections_sel = " + str(startProj) + ":" +str(endProj) + "\n")
        myFile.write("acqui_inv = " + str(invertAcqui) + "\n")
        myFile.write('\n')


def addDetectorInfo(f, detWidth, detX, detY, detZ):
    with open(f, 'a') as myFile:
        myFile.write("[detector_setup]\n")
        myFile.write("det_pos_x = " + str(detWidth) + "\n")
        myFile.write("det_pos_y = " + str(detX) + "\n")
        myFile.write("det_pos_z = " + str(detY) + "\n")
        myFile.write("det_width = " + str(detY) + "\n")
        myFile.write('\n')



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/
"""

"""

__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "18/10/2017"

import os
from . import structs
from . import config as configmod


class _ConfigIO():
    """Class able to read and write configuration"""
    VALID_FILE_EXTENSIONS = None

    def isAValidFileName(self, filePath):
        fn, file_extension = os.path.splitext(filePath)
        return file_extension in self.VALID_FILE_EXTENSIONS


class ConfigReader(_ConfigIO):
    """Base class of the configuration reader"""
    dictReconsTypeToReconsConf = {
        configmod._ReconsConfig.TX_ID: configmod.TxConfig,
        configmod._ReconsConfig.FLUO_ID: configmod.FluoConfig,
        configmod._ReconsConfig.COMPTON_ID: configmod.FluoConfig
    }
    """Dictionary used to associate a reconstruction type and a configuration
    class"""

    def read(self, filePath):
        """

        :param filePath: path of the file to read
        :return: an instance of configuration. Might be FluoConfiguration or
                 TxConfiguration depending on the file read.
        """
        raise NotImplementedError("Config is an abstract class")

    def _reloadDatasets(self, config):
        """Load all dataset contained in the configuration"""
        self._loadI0(config)
        self._loadSinograms(config)
        if isinstance(config, configmod.FluoConfig):
            self._loadAbsMat(config)
            self._loadMaterials(config)
        return config

    def _loadSinograms(self, config):
        for sinogram in config.sinograms:
            assert(isinstance(sinogram, structs.DataStored))
            sinogram.loadData()
            if hasattr(sinogram, 'selfAbsMat') and sinogram.selfAbsMat:
                sinogram.selfAbsMat.loadData()

    def _loadAbsMat(self, config):
        if config.absMat is not None:
            assert isinstance(config.absMat, structs.AbsMatrix)
            config.absMat.loadData()

    def _loadMaterials(self, config):
        if config.materials:
            assert isinstance(config.materials, structs.Materials)
            config.materials.loadData()

    def _loadI0(self, config):
        if isinstance(config.I0, structs.DataStored):
            config.I0.loadData()


class ConfigWriter(_ConfigIO):
    """Class to write a configuration into a config file
    
    :param filePath: output file
    :param :class:`_ReconsConfig` reconsConfiguration: configuration to write
    :param bool overwrite: if True overwrite contains if file exists
    :param bool merge: include all the dataset of the
                                      configuration in the file.
    """
    def write(self, filePath, reconsConfiguration, overwrite=False,
              merge=False):
        """"""
        raise NotImplementedError("Config is an abstract class")



  #!/usr/bin/env python
# coding: utf-8
# /*##########################################################################
# Copyright (C) 2016 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ############################################################################*/
"""
Simple application doing data.max() - data
"""

__authors__ = ["H> Payno"]
__license__ = "MIT"
__date__ = "13/08/2018"

from freeart.utils import (reconstrutils, outgoingrayalgorithm, raypointsmethod)
import freeart
from freeart.unitsystem import metricsystem
import argparse
import sys
import logging
import numpy


def getinputinfo():
    return "freeart reconsfluosino sample.edf absmat.edf selfasbmat.edf [nangles]"


def main(argv):
    parser = argparse.ArgumentParser(description=__doc__)
    parser.add_argument('sinogramfile',
                        help='file containing the fluorescence sinogram')
    parser.add_argument('absmatfile',
                        help='definition of the sample absorption (in g.cm-2)')
    parser.add_argument('selfabsmatfile',
                        help='definition of the sample self attenuation (in g.cm-2)')
    parser.add_argument('--output', '-o',
                        dest="output",
                        default=None,
                        help='outputfile')
    parser.add_argument('--I0', '--i0',
                        dest="i0",
                        type=float,
                        default=1.0)
    parser.add_argument('--voxelsize',
                        dest="voxelsize",
                        type=float,
                        default=1.0*metricsystem.centimeter)
    parser.add_argument('--nangles',
                        dest="nangles",
                        type=int,
                        default=None,
                        help='number of projections')
    parser.add_argument('--halfacq',
                        dest="halfacq",
                        action="store_true",
                        default=False,
                        help='Do we want to make a sinogram from 0 to 180 or from 0 to 360 degree')
    parser.add_argument('--niter',
                        dest="niter",
                        type=int,
                        default=4,
                        help='Number of iteration to process')
    parser.add_argument('--relaxation_factor',
                        dest="relaxation_factor",
                        default=None,
                        help='Relaxation factor to apply to the ART algorithm')
    parser.add_argument('--debug',
                        dest="debug",
                        action="store_true",
                        default=False,
                        help='Set logging system in debug mode')
    options = parser.parse_args(argv[1:])
    if options.debug:
        logging.root.setLevel(logging.DEBUG)

    _halfacq = options.halfacq

    _output = options.output
    if _output and _output.endswith('.edf') is False:
        _output = _output + '.edf'

    sinogram_file = options.sinogramfile
    assert sinogram_file.lower().endswith('.edf')
    absmat_file = options.absmatfile
    assert absmat_file.lower().endswith('.edf')
    selfabsmat_file = options.selfabsmatfile
    assert selfabsmat_file.lower().endswith('.edf')

    sinogram_data = reconstrutils.LoadEdf_2D(sinogram_file)
    sinogram_data = reconstrutils.increaseMatSize(sinogram_data, True)

    absmat_data = reconstrutils.LoadEdf_2D(absmat_file)
    absmat_data = reconstrutils.increaseMatSize(absmat_data, False)

    selfabsmat_data = reconstrutils.LoadEdf_2D(selfabsmat_file)
    selfabsmat_data = reconstrutils.increaseMatSize(selfabsmat_data, False)

    relaxation_factor = options.relaxation_factor
    if relaxation_factor is None:
        relaxation_factor = 1.0 / float(sinogram_data.shape[1])
    else:
        relaxation_factor = float(relaxation_factor)

    detPos = (1000 * metricsystem.cm, 0, 0)  # 90 degree detector
    detectorRadius = 1 * metricsystem.cm
    detSetup = [(detPos, detectorRadius)]
    oversampling = 2
    maxAngle = numpy.pi * 2.0 if _halfacq is False else numpy.pi
    angles = numpy.linspace(0.0, maxAngle, sinogram_data.shape[1])

    print('----- configuration ---')
    print('sinogram file: ', sinogram_file)
    print('absorption matrix file: ', absmat_file)
    print('self absorption mattrix file: ', selfabsmat_file)
    print('half acq: ', _halfacq)
    print('oversampling: ', oversampling)
    print('I0: ', options.i0)
    print('relaxation factor: ', relaxation_factor)
    print('nb iteration: ', options.niter)
    print('angles, from: 0 to ', maxAngle, ' nb angles: ', str(sinogram_data.shape[1]))
    print('-----------------------')
    print('start reconstruction')

    # TODO: add possibility to define the voxel size in the command line

    # run the sinogram generation
    alRecons = freeart.FluoBckProjection(sinoDat=sinogram_data,
                                         sinoAngles=angles,
                                         expSetUp=detSetup,
                                         absorp=absmat_data,
                                         selfAbsorp=selfabsmat_data)

    alRecons.setOverSampling(oversampling)
    alRecons.setDampingFactor(relaxation_factor)
    # alRecons.setRandSeedToZero(True)
    alRecons.setVoxelSize(options.voxelsize)
    alRecons.turnOffSolidAngle(True)
    alRecons.setRayPointCalculationMethod(raypointsmethod.withInterpolation)
    alRecons.setOutgoingRayAlgorithm(outgoingrayalgorithm.rawApproximation)

    phantom = alRecons.iterate(options.niter)

    if _output is not None:
        reconstrutils.saveMatrix(data=phantom, fileName=_output, overwrite=True)
        # TODO: else plot it using silx
    print('end sinogram generation')


if __name__ == "__main__":
    """
    """
    main(sys.argv)



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/

__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "25/09/2017"

import re
import numpy

SELECTION_PATTERN = re.compile("\d?[:]?\d?")


def selectionIsValid(selection):
    """
    Return true if the given selection as a string is valid

    :return: bool
    """
    assert type(selection) is str
    _selection = selection.replace(" ", "")
    selections = _selection.split(";")
    for sel in selections:
        if not re.match(SELECTION_PATTERN, sel):
            return False
    return True


def getSelection(projections, selection):
    """

    :param str selection: 
    :return numpy.ndarray: 
    """
    def evalSelection(projections, sel):
        assert(type(projections) is numpy.ndarray)
        return eval('projections[' + sel + ']')
    assert projections.ndim in (1, 2)

    definition = None
    if projections.ndim is 2:
        definition = projections.shape[1]
    assert type(selection) is str
    if selection == '' or selection == 'None' or selection is None:
        return projections
    _selection = selection.replace(" ", "")
    selections = _selection.split(";")
    if len(selections) is 1:
        res = evalSelection(projections, selection)
    else:
        res = None
        for iSel, sel in enumerate(range(len(selections))):
            if iSel is 0:
                res = evalSelection(projections, selections[iSel])
            else:
                res = numpy.append(res,
                                   evalSelection(projections,
                                   selections[iSel]))

    if definition is not None:
        res = res.reshape(-1, definition)
    return res


def getAngles(minAngle, maxAngle, nbAngles, selection, lastAngleEqualFirst,
              acquiInverted):
    """
    
    :param float minAngle: angle of the first projection in radian
    :param float maxAngle: angle of the last projection in radian 
    :param int nbAngles: number of projection angle contained in the sinogram
    :param str selection: the selection of the angle to consider
    :param bool lastAngleEqualFirst: True if maxAngle is the value of the last
                                     projection or if it has to be removed.
    :param bool acquiInverted: True if the acquisition have been inverted and
                               angles have to be inverted to. 
    :return numpy.ndarray: the angles fitting the description
    """
    assert(type(lastAngleEqualFirst) is bool)
    assert(type(acquiInverted) is bool)
    angles = numpy.linspace(minAngle,
                            maxAngle,
                            nbAngles if lastAngleEqualFirst is True else nbAngles + 1)
    if acquiInverted is True:
        angles = angles * (-1.0)
    if lastAngleEqualFirst is False:
        angles = angles[:-1]

    assert(len(angles) == nbAngles)
    return getSelection(angles, selection)



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/
__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "01/09/2016"

import unittest
from unittest.util import safe_repr

class testPercentage(unittest.TestCase): 
    """A class to make test from percentage of error between elemets.
    Will redefine some assert from unittest.TestCase"""

    def __init__(self, methodName='runTest'):
        """Create an instance of the class that will use the named test
           method when executed. Raises a ValueError if the instance does
           not have a method with the specified name.
        """
        self._testMethodName = methodName
        self._outcome = None
        self._testMethodDoc = 'No test'
        try:
            testMethod = getattr(self, methodName)
        except AttributeError:
            if methodName != 'runTest':
                # we allow instantiation with no explicit method name
                # but not an *incorrect* or missing method name
                raise ValueError("no such test method in %s: %s" %
                      (self.__class__, methodName))
        else:
            self._testMethodDoc = testMethod.__doc__
        self._cleanups = []
        self._subtest = None

    def getError(self, first, second) :
        meanVal = (first + second) /2.0
        return (abs(first - second) / meanVal) * 100.0

    def assertErrorPercentageAlmostEqual(self, first, second, acceptableErrorPercentage) :
        """Fail if the percentage of the difference between first and second is higher than acceptableErrorPercentage"""
        if(first == second) :
            # shortcut
            return

        error = self.getError(first, second)

        if(error > acceptableErrorPercentage ) :
            msg = '%s differ from %s with more than %s %% : %s %%' % (safe_repr(first),
                                                          safe_repr(second),
                                                          acceptableErrorPercentage,
                                                          error)
            raise self.failureException(msg)
        else :
            return


# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/
"""
Python module to build and run a freeart reconstruction from a configuraton
file (.cfg file).
"""

__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "23/05/2016"


from freeart.configuration import read as readConfig
from .configinterpreter import TxConfInterpreter, FluoConfInterpreter
import freeart.utils.reconstrutils as reconstrutils
from threading import Thread
import os
import logging

_logger = logging.getLogger(__name__)


class GlobalConfigInterpreter():
    """
    This is hte main class of the module.
    The Global interpreter will call other interpreters in order to interpret a
    configuration file into a transmission or a fluorescence reconstruction.

    :param filePath: the file path of th ecfg file
    :param config: the `_ReconsConfig` if already exists
    :param force: if True, will force the interpreter to be created even if the
                  cfg file is invalid. This parameter is used for example in
                  the case that the we have a fluorescence reconstruction and
                  that we want to build a transmission reconstruction first
    """

    interpreter = None
    """The interpreter used to convert a .cfg file into a freeart
    reconstruction"""

    def __init__(self, filePath, config=None, force=False):
        self.filePath = filePath
        if self.filePath is not None:
            self._checkFileIsCorrect()

        if config is None:
            self._config = readConfig(self.filePath)
        else:
            self._config = config

        if self.isATxReconstruction():
            inf = """Transmission reconstruction file detected. Creating the
                reconstruction algorithm"""
            _logger.info(inf)
            self.interpreter = TxConfInterpreter(filePath, self._config)
        elif self._config.isAbsMatASinogram is True:
            inf = """In order to run the fluorescence reconstruction the
             absorption matrix need to be reconstructed first. Please run
             this reconstruction first.
             """
            raise ValueError(inf)
        elif self.isAFluoOrComptonReconstruction():
            inf = """Fluorescence reconstruction file detected. Creating the
                reconstruction algorithm"""
            _logger.info(inf)
            self.interpreter = FluoConfInterpreter(filePath, self._config)

    def getConfig(self):
        return self.interpreter.config

    config = property(getConfig)

    def getReconstructionAlgorithms(self):
        """
        Return one ART algorithm for each sinogram we want to reconstruct
        """
        if self.interpreter:
            return self.interpreter.getReconstructionAlgorithms()
        else:
            inf = "No interpreter found during construction for the given file"
            _logger.info(inf)
            return None

    def isATxReconstruction(self):
        """

        :return: True if the given file is set up for a transmission
        reconstruction"""
        return self._config.reconsType == 'Transmission'

    def isAFluoOrComptonReconstruction(self):
        """

        :return: True if the given file is set up for a fluorescence or
        compton reconstruction"""
        return self._config.reconsType in ('Fluorescence', 'Compton')

    def _checkFileIsCorrect(self):
        """chack that the givem cfg file is correct"""
        if not type(self.filePath) is str:
            err = "%s should be a string" % self.filePath
            raise TypeError(err)
        if not os.path.isfile(self.filePath):
            err = "%s should be an existing file" % self.filePath
            raise ValueError(err)

    def iterate(self, nbIteration):
        """
        Run nb Iteration, if possible in parallel

        :param nbIteration: the number of iteration we want to run
            (one iteration will execute a backward projection from each angle
                of acquisition)
        """
        t = Thread(target=self.interpreter.iterate, args=(nbIteration,))

        t.start()
        t.join()

    def setRandSeedToZero(self, b):
        self.interpreter.setRandSeedToZero(True)

    def saveCurrentReconstructionsTo(self, folderPath):
        """
        save one file per reconstruction to the given folder

        :param folderPath: the path we want to save under
        """
        assert(self.interpreter is not None)
        assert(os.path.isdir(folderPath))
        folderPath = folderPath + "/"
        algos = self.getReconstructionAlgorithms()
        for algoName in algos:
            outputFile = folderPath + 'recons_' + algoName
            if not outputFile.lower().endswith('.edf'):
                outputFile = outputFile + '.edf'
            reconstrutils.savePhantom(algos[algoName].getPhantom(), outputFile)

    def setOversampling(self, val):
        """
        Set the given oversampling to all the ART algorithms of the interpreter

        :param val: the new oversampling value.
        """
        assert(type(val) == int)
        assert(type(val) > 0)
        ARTAglos = self.interpreter.getReconstructionAlgorithms()
        for algo in ARTAglos:
            ARTAglos[algo].setOverSampling(val)

    def setDampingFactor(self, val):
        """
        Set the given dampingFactor to all the ART algorithms of the
        interpreter

        :param val: the new dampingFactor value.
        """
        assert(type(val) == float)
        assert(val > 0)
        assert(val <= 1.0)
        ARTAglos = self.interpreter.getReconstructionAlgorithms()
        for algo in ARTAglos:
            ARTAglos[algo].setDampingFactor(val)

    def setRayPointCalculationMethod(self, methodID):
        """
        Set the given dampingFactor to all the ART algorithms of the
        interpreter

        :param methodID: the new method to compute the ray points values
            (with or without interpolation).
        """
        ARTAglos = self.interpreter.getReconstructionAlgorithms()
        for algo in ARTAglos:
            ARTAglos[algo].setRayPointCalculationMethod(methodID)

    def setOutgoingRayAlgorithm(self, algorithmID):
        """
        Set the given dampingFactor to all the ART algorithms of the
        interpreter

        :param algorithmID: the new algorithmID value.
        """
        ARTAglos = self.interpreter.getReconstructionAlgorithms()
        for algo in ARTAglos:
            ARTAglos[algo].setOutgoingRayAlgorithm(algorithmID)

    def getReconstructionType(self):
        """

        :return: the type of reconstruction we want to run
        """
        return self.interpreter.config.reconsType

    def getE0(self):
        """

        :return: the energy of the incoming rays. Units of Energy is keV
        """
        return self.config.E0

    def setE0(self, E0):
        """
        Set the energy of the incoming rays. This is needed for fluorescence
        reconstruction.

        :param E0: the energy of the incoming rays. In keV
        """
        self.config.E0 = E0



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/

"""

..warning :: those example have been created before the existance of the
             :class: Txconfig and :class: configuration.config.FluoConfig
             classes.
             Due to lake of time I couldn't rewrote them but they are certainly
             not an easy access to understand the configuration and how a
             reconstruction is created even if the name migh tlet understand
             differently.
             If you have a configuration to write you better look at the
             configuration.config module.
"""
__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "01/09/2016"


import os
import tempfile

import fabio
import numpy as np
from silx.io import configdict
import freeart
from freeart.configuration import iniConfigIO, config, structs, fileInfo
from freeart.interpreter import configinterpreter
from freeart.unitsystem import metricsystem
from freeart.utils import getDistance, getDistance_2D, getAngle
from freeart.utils import physicalelmts, genph, reconstrutils, raypointsmethod, \
    outgoingrayalgorithm
from freeart.utils import h5utils
from silx.io import configdict
from collections import OrderedDict
from silx.io import dictdump
from silx.io.url import DataUrl


def toExperimentationRef(pt, phDimx, phDimy, phDimz):
    """
    Transform the point pt to the FreeART reference for the according giving
    phantom dimesions

    ..Note:: For freeART the center of the universe is the medium voxel
    """
    return (pt[0]-phDimx//2, pt[1]-phDimy//2, pt[2]-phDimz//2)


def getAngleBetweenVec_DetectorVoxel_Vec_x(detPos, voxPos):
    """
    Return the angle made between the vector detPosVoxPos and the X vector.
    Which can be see as the vector of each voxel. And be used to know the
    angle of incidence of a ray with the voxel.
    """
    vectorX = (1.0, 0.0, 0.0)
    vectorDetVox = (detPos[0] - voxPos[0], detPos[1] - voxPos[1], detPos[2] - voxPos[2])
    return getAngle(vectorX, vectorDetVox)


def getSinAngleDetectorVoxel(detPos, voxPos):
    """Return the sinus of the angle between detPos and voxPos"""
    return np.sin(getAngleBetweenVec_DetectorVoxel_Vec_x(detPos, voxPos))


def getSolidAngle(voxelPosInFreeARTRef, detPos, detectorRadius):
    """
    Compute detector efficiency according to detector position and voxel
    position
    """
    distDetectorPoint = getDistance(voxelPosInFreeARTRef, detPos)
    # get the angle between the vectors (detcent, origin) and (voxelpos, origin)
    cosAlpha = distDetectorPoint / np.sqrt(distDetectorPoint*distDetectorPoint + detectorRadius*detectorRadius)
    return 0.5*(1-cosAlpha)


def getRawSumTheoreticalFluoValue(detPos, detectorRadius, voxelPosInFreeARTRef,
                                  mu_E0, mu_EF, psiIncoming, psiOutgoing,
                                  widths, I0=1.0):

    signalSum = 0
    for lWidth in widths:
        signalSum += getRawTheoreticalFluoValue(detPos=detPos,
                                                detectorRadius=detectorRadius,
                                                voxelPosInFreeARTRef=voxelPosInFreeARTRef,
                                                mu_E0=mu_E0,
                                                mu_EF=mu_EF,
                                                psiIncoming=psiIncoming,
                                                psiOutgoing=psiOutgoing,
                                                material_width=lWidth+1.0,
                                                I0=I0)

    return signalSum


def getRawTheoreticalFluoValue(detPos, detectorRadius, voxelPosInFreeARTRef,
                               mu_E0, mu_EF, psiIncoming, psiOutgoing,
                               material_width, I0=1.0):

    detectorEfficiency = getSolidAngle(voxelPosInFreeARTRef,
                                       detPos,
                                       detectorRadius)

    sinPsiIncoming = np.sin(psiIncoming)
    sinPsiOutgoing = np.sin(psiOutgoing)

    incomingBeamAttenuation = np.exp(-mu_E0*material_width / sinPsiIncoming)
    outgoingBeamAttenuation = np.exp(-mu_EF*material_width / sinPsiOutgoing)
    return I0 * incomingBeamAttenuation * outgoingBeamAttenuation * detectorEfficiency


def getFluoTheoreticalValue(detPos, detectorRadius, voxelPosInFreeARTRef,
                            mu_E0, mu_EF, psiIncoming, psiOutgoing, tau_E0,
                            distInMat, I0=1.0, Ci=1.0, Wis=1.0, Ris_EF=1.0):
    """
    Return the theoretical I value for a voxel according to fluorescence theory
    The intensity get for the Energy EF
    We are looking for fluoresence photon emitted by the element i
    This work for a beam and an homogeneous material.

    :param detPos : The position of the detector
    :param voxelPosInFreeARTRef : the position of the computed voxel in the
                                  freeart Reference
    :param mu_E0: the mass attenuation of the incoming beam at energy E0
    :param mu_EF: the mass attenuation of the outgoing beam at energy EF
    :param psiIncoming: the angle of entry of the incoming beam in the material
    :param psiOutgoing: the conjuge of the angle between the incoming beam and
                        the voxel-detector vector
    :param tau_E0: Probability to create a vacancy
    :param I0 : Intensity of the incoming beam
    :param Ci : mass fraction of the element i
    :param Wis: probqbility for beeing in the atomic shell s ?
    :param Ris_EF : Probability of generating an element i ?
    """

    detectorEfficiency = getSolidAngle(voxelPosInFreeARTRef,
                                       detPos,
                                       detectorRadius)

    # Compute the theoretical value
    proba_emitting_photon_element_i = Ci*Wis*Ris_EF

    # if no absorption
    if (mu_E0 == 0) and (mu_EF == 0):
        return proba_emitting_photon_element_i*detectorEfficiency

    sinPsiIncoming = np.sin(psiIncoming)
    sinPsiOutgoing = np.sin(psiOutgoing)

    Fact_1 = tau_E0 / (mu_E0 + mu_EF*sinPsiIncoming/sinPsiOutgoing)
    expCalculation = -(mu_E0/sinPsiIncoming + mu_EF/sinPsiOutgoing) * distInMat
    return I0 * proba_emitting_photon_element_i * Fact_1 * (1.0 - np.exp(expCalculation)) * detectorEfficiency


def getFluoTheoreticalValueForTransmission(detPos, detectorRadius,
                                           voxelPosInFreeARTRef, mu_E0, mu_EF,
                                           psiIncoming, psiOutgoing, tau_E0,
                                           distInMat, I0=1.0, Ci=1.0, Wis=1.0,
                                           Ris_EF=1.0):
    """
    Return the theoretical I value for a voxel according to fluorescence theory

    ..Warning:: : we made the assumption that rays are parallel.
    We are looking for fluoresence photon emitted by the element i
    This work for a beam and an homogeneous material.

    :param detPos : The position of the detector
    :param voxelPosInFreeARTRef : the position of the computed voxel in the
                                  freeart reference
    :param mu_E0: the mass attenuation of the incoming beam at energy E0
    :param mu_EF: the mass attenuation of the outgoing beam at energy EF
    :param psiIncoming: the angle of entry of the incoming beam in the material
    :param psiOutgoing: the conjuge of the angle between the incoming beam and
                        the voxel-detector vector
    :param tau_E0: Probability to create a vacancy
    :param I0 : Intensity of the incoming beam
    :param Ci : mass fraction of the element i
    :param Wis: probability for beeing in the atomic shell s ?
    :param Ris_EF : Probability of generating an element i ?
    """

    detectorEfficiency = getSolidAngle(voxelPosInFreeARTRef, detPos, detectorRadius)

    proba_emitting_photon_element_i = Ci*Wis*Ris_EF

    sinPsiIncoming = np.sin(psiIncoming)
    sinPsiOutgoing = np.sin(psiOutgoing)

    Fact_1 = tau_E0 / (mu_E0 - mu_EF*sinPsiIncoming/sinPsiOutgoing)
    expCalculationE = -(mu_E0/sinPsiIncoming) * distInMat
    expCalculationF = -(mu_EF/sinPsiOutgoing) * distInMat

    return I0 * proba_emitting_photon_element_i * Fact_1 * \
           (np.exp(expCalculationF) - np.exp(expCalculationE)) * detectorEfficiency


def removeAvoidVoxel(phantom, replacementValue=0.0):
    """
    Set all voxel which aren't in the higher circle contained by the phantom to
    zero.
    This function can be used to ignore voxel which were not in the area
    treated by the freeART reconstruction algorithm.
    """
    circleCenter = (phantom.shape[0]//2, phantom.shape[1]//2)
    circleRadius = min(phantom.shape[0], phantom.shape[1])//2
    for z in np.arange(0, phantom.shape[2], 1):
        for x in np.arange(0, phantom.shape[0], 1):
            for y in np.arange(0, phantom.shape[1], 1):
                if getDistance_2D(circleCenter, (x, y)) > circleRadius:
                    phantom[x, y, z] = replacementValue

    return phantom


class SheppLoganReconstruction():
    """
    Test of the generation of a sinogram and of the reconstruction of it
    """
    methodGBeams = raypointsmethod.withInterpolation

    optionOutgoing = outgoingrayalgorithm.createOneRayPerSamplePoint

    def produce_SheppLogan_sinogram(self, dim, _oversampling, _anglesNb,
                                    _detSetup,
                                    _voxelSize=1.0*metricsystem.centimeter):
        """
        Example of the generation of a simple 2D sinogram.
        :param _oversampling: the number of oversampling number we want o take
                              per mm covered by the ray
        :param _anglesNb: the number of angle we want to make an acquisition
                          for.
        """
        # generate the phantom
        phGenerator = genph.PhantomGenerator()
        sheppLogan_phantom = phGenerator.get2DPhantomSheppLogan(dim)
        sheppLogan_phantom.shape = (sheppLogan_phantom.shape[0],
                                    sheppLogan_phantom.shape[1],
                                    1)

        # generate absMat and selfAnsMat
        absMat = sheppLogan_phantom * 20.0
        selfAbsMat = sheppLogan_phantom / 10.0

        # run the sinogram generation
        sinogram, angles = reconstrutils.makeFreeARTFluoSinogram(
            sheppLogan_phantom,
            absMat,
            selfAbsMat,
            _anglesNb,
            _detSetup,
            _oversampling,
            self.methodGBeams,
            self.optionOutgoing,
            voxelSize=_voxelSize)

        return sinogram, angles, absMat, selfAbsMat, sheppLogan_phantom

    def make_reconstruction(self, _sinogram, _angles, _absMat, _selfAbsMat,
                            _detSetup, _nbIter, _oversampling,
                            _voxelSize=1.0*metricsystem.cm,
                            _dampingFactor=0.005, _turnOffSolidAngle=False):
        """
        Example of the reconstruction from a sinogram and related data.

        :param _sinogram: the sinogram to use for the reconstruction
        :param _angles: the angles of acquisition registred on the sinogram
        :param _nbIter: the number of iteration to execute for the
                       reconstruction
        """

        alRecons = freeart.FluoBckProjection(sinoDat=_sinogram,
                                             sinoAngles=_angles,
                                             expSetUp=_detSetup,
                                             absorp=_absMat,
                                             selfAbsorp=_selfAbsMat)

        alRecons.setOverSampling(_oversampling)
        alRecons.setDampingFactor(_dampingFactor)
        alRecons.setRayPointCalculationMethod(self.methodGBeams)
        alRecons.setOutgoingRayAlgorithm(self.optionOutgoing)
        # alRecons.setRandSeedToZero(True)
        alRecons.setVoxelSize(_voxelSize)
        alRecons.turnOffSolidAngle(_turnOffSolidAngle)

        return alRecons.iterate(_nbIter)


class FluorescenceExperimentationExample(object):
    """
    A simple fluorescence experimentation example

    :param buildFromSinogram: if True then we will build the absorption matrix
                              from a sinogram
    :param elements: the list of element for which we want to generate a
                     sinogram
    :param materials: The list of the materials to consider
    :param sheppLoganPartIDs: Dictionnary that for each material attribute an
                              area of the shepplogan
    :param densities: For each material the density of the material.
                      Should be between 0 and 1
    :param oversampling: the oversampling per step to run
    :param dampingFactor: the relaxation factor to apply
    :param nbAngles: the number of projection
    :param buildFromSinogram: if True then we will set the sinogram file in
                              the config file
    :param E0:
    :param EF:
    :param matrixWidth:
    :param minAngle: the angle of the last projection.
    :param maxAngle: the angle of the last projection.
    :param firstProjection: the index of the first projection to include for
                            the reconstruction.
    :param lastProjection: the index of the last projection to include for the
                           reconsutruction.
    :param setProbaEmissionToOne: if True, skip the call to fisx to compute the
                                  probaility of emission
    """
    voxelSize = 1.0 * metricsystem.cm
    detectorRadius = 1.0 * metricsystem.cm
    det = structs.Detector(x=0., y=1000.0 * metricsystem.cm, z=0.,
                           width=1.0 * metricsystem.cm)
    detPos = (0.0, 1000.0*metricsystem.cm, 0.0)
    detSetup = [(detPos, detectorRadius)]

    def __init__(self, elements, materials, sheppLoganPartIDs, densities,
                 oversampling, dampingFactor, EF, nbAngles=360,
                 buildFromSinogram=False, E0=1.0e6, matrixWidth=56,
                 minAngle=0.0, maxAngle=2.0*np.pi, firstProjection=0,
                 lastProjection=None, includeLastProjection=False,
                 setProbaEmissionToOne=False, turnOffSolidAngle=False):

        # check given densitites and interaction probabilities
        for mat in materials:
            if mat not in densities:
                err = 'Density is not given for %s' % mat
                raise ValueError(err)
            if densities[mat] < 0 or densities[mat] > 1:
                err = 'Density for %s is not in [0, 1], non sense' % mat
                raise ValueError(err)

        self.materials = materials
        self.elements = elements
        self.densities = densities
        self.E0 = E0
        self.EF = EF
        self._buildFromSinogram = buildFromSinogram
        self.minAngle = minAngle
        self.maxAngle = maxAngle
        self.firstProjection = firstProjection
        self.lastProjection = lastProjection
        self.includeLastProjection = includeLastProjection
        self.setProbaEmissionToOne = setProbaEmissionToOne

        self.sheppLoganPartIDs = sheppLoganPartIDs

        self.selfAbsorptionMatrices = {}
        self.sinos = {}
        self.phantoms = {}
        self.oversampling = oversampling
        self.dampingFactor = dampingFactor
        self.nbAngles = nbAngles
        self.matWidth = matrixWidth
        self.turnOffSolidAngle = turnOffSolidAngle

    def setUp(self):
        """
        Main function of the setUp
        """
        self.tempdir = tempfile.mkdtemp()
        self.setUpMaterialsIntoFisx()
        self.setUpElementsMatrice()
        self.setUpDensityMatrice()
        self.setUpAbsMat()
        self.setUpInteractionMatrice(self.setProbaEmissionToOne)
        self.setUpSelfAbs()
        sinosFilesInfo = self.setUpFluorescenceSinograms()
        conf = self.setUpConfigFile(sinosFilesInfo)
        self.writeConfigFile(conf)
        self.outputdir = tempfile.mkdtemp()

    def setUpMaterialsIntoFisx(self):
        """Register all needed material into fisx"""
        import fisx
        from fisx import Elements
        self.fisxElements = Elements()
        self.fisxMaterials = {}

        for material in self.materials:
            mat = fisx.Material(material,
                                self.materials[material]['Density'], self.materials[material]['Thickness'])
            mat.setComposition(
                dict(zip(self.materials[material]['CompoundList'], self.materials[material]['CompoundFraction'])))
            self.fisxElements.addMaterial(mat)
            self.fisxMaterials[material] = mat

    def setUpElementsMatrice(self):
        """
        Create the matrice for the sample that set for each voxel the material

        ..Note:: a voxel can contain only one material
        """
        phGenerator = genph.PhantomGenerator()
        self.matOfPhysElmts = np.ndarray(shape=(self.matWidth, self.matWidth),
                                         dtype='S10')
        self.matOfPhysElmts[:] = ''
        # basic behavior : a voxel is limited to one material
        for materialName in self.materials:
            partMaterial = phGenerator.get2DPhantomSheppLogan(self.matWidth,
                                                              self.sheppLoganPartIDs[materialName])
            self.matOfPhysElmts[partMaterial != 0.0] = materialName

    def setUpDensityMatrice(self):
        """
        Create the density matrix (material densities)
        """
        self.densityMat = np.zeros(shape=(self.matWidth, self.matWidth), dtype='float64')
        # basic behavior : a voxel is limited to one material
        for materialName in self.materials:
            condition = np.in1d(self.matOfPhysElmts.ravel(), materialName).reshape(self.matOfPhysElmts.shape)
            self.densityMat[condition] = self.densities[materialName]

        self.densityMat.shape = (self.densityMat.shape[0],
                                 self.densityMat.shape[1],
                                 1)

    def setUpAbsMat(self):
        """
        Generate the absorption matrix (absorption of the incoming beams)
        """
        self.absorption_matrix_file = os.path.join(self.tempdir,
                                                   "absorption_matrix.edf")
        self.absorption_matrix_value = np.zeros(shape=(self.matWidth, self.matWidth),
                                                dtype=np.float64)
        self.absMat = structs.AbsMatrix(data=self.absorption_matrix_value,
                                        fileInfo=fileInfo.MatrixFileInfo(file_path=self.absorption_matrix_file))

        physicalelmts.elementInstance.removeMaterials()

        for materialName in self.materials:
            physicalelmts.elementInstance.addMaterial(reconstrutils.convertToFisxMaterial(materialName,
                                                                                          self.materials[materialName]))

            absValue = physicalelmts.elementInstance.getElementMassAttenuationCoefficients(materialName,
                                                                                           self.E0)
            condition = np.in1d(self.matOfPhysElmts.ravel(), materialName).reshape(self.absorption_matrix_value.shape)
            self.absorption_matrix_value[condition] = absValue['total']

        if self._buildFromSinogram:
            self.absorption_matrix_value.shape = (self.absorption_matrix_value.shape[0],
                                                  self.absorption_matrix_value.shape[1],
                                                  1)
            al = freeart.TxFwdProjection(self.absorption_matrix_value,
                                         minAngle=self.minAngle,
                                         maxAngle=self.maxAngle,
                                         anglesNb=self.nbAngles)
            al.setOverSampling(5)
            sinogram, angles = al.makeSinogram()
            sinogram.shape = (sinogram.shape[1], sinogram.shape[2])
            reconstrutils.saveSinogram(sinogram, self.absorption_matrix_file)
            self.sinogramTx = sinogram
        else:
            reconstrutils.saveMatrix(self.absorption_matrix_value,
                                     self.absorption_matrix_file)
            self.absorption_matrix_value.shape = (self.absorption_matrix_value.shape[0],
                                                  self.absorption_matrix_value.shape[1],
                                                  1)

    def setUpInteractionMatrice(self, setProbaEmissionToOne):
        """
        This will compute the interaction matrice.
        This is for each voxel the probability of emission of the element for
        the material multiply by the density of the material

        .. Note: density of the each material is suppossed homogeneous in the
                 sample.

        :param setProbaEmissionToOne: if True, skip the call to fisx to compute
                                      the probaility of emission
        """
        self.interaction_matrix_file = {}
        self.interaction_matrix_value = {}
        for element in self.elements:
            fName = "interaction_matrix_" + element + ".edf"
            self.interaction_matrix_file[element] = os.path.join(self.tempdir, fName)
            self.interaction_matrix_value[element] = np.zeros(shape=(self.matWidth, self.matWidth),
                                                              dtype=np.float64)

            for materialName in self.materials:
                if setProbaEmissionToOne is False:
                    probabilityOfEmission = configinterpreter.FluoConfInterpreter.getProbabilityOfEmission(fisxMat=self.fisxMaterials[materialName],
                                                                                                           energy=self.E0,
                                                                                                           element=element,
                                                                                                           fisxElements=self.fisxElements,
                                                                                                           shell=None)
                else:
                    probabilityOfEmission = 1.0

                condition = np.in1d(self.matOfPhysElmts.ravel(), materialName).reshape(self.interaction_matrix_value[element].shape)
                self.interaction_matrix_value[element][condition] = probabilityOfEmission

            reconstrutils.saveMatrix(self.interaction_matrix_value[element],
                                     self.interaction_matrix_file[element])

            s0 = self.interaction_matrix_value[element].shape[0]
            s1 = self.interaction_matrix_value[element].shape[1]
            self.interaction_matrix_value[element].shape = (s0, s1, 1)

    def setUpFluorescenceSinograms(self):
        """
        Produce fluorescence sinograms
        """
        def buildElementSelfAbsMat(element):
            selfAbsMat = np.zeros(shape=(self.matWidth, self.matWidth),
                                  dtype=np.float64)
            for materialName in self.materials:
                condition = np.in1d(self.matOfPhysElmts.ravel(), materialName).reshape(selfAbsMat.shape)
                selfAbsMat[condition] = physicalelmts.elementInstance.getElementMassAttenuationCoefficients(
                    materialName, self.EF[element])['total'][0]

            selfAbsMat.shape = (selfAbsMat.shape[0], selfAbsMat.shape[1], 1)

            return selfAbsMat

        self.sinogramsdir = tempfile.mkdtemp()
        self.sino_file = os.path.join(self.sinogramsdir, "sinograms_fluo.edf")

        for element in self.elements:
            ph_value = self.interaction_matrix_value[element]*self.densityMat
            selfAbsMat = buildElementSelfAbsMat(element)

            sino_value = reconstrutils.makeFreeARTFluoSinogram(ph_value,
                                                               self.absorption_matrix_value,
                                                               selfAbsMat,
                                                               minAngle=self.minAngle,
                                                               maxAngle=self.maxAngle,
                                                               numAngle=self.nbAngles,
                                                               detSetup=self.detSetup,
                                                               oversampling=self.oversampling,
                                                               beamCalcMeth=raypointsmethod.withInterpolation,
                                                               outRayPtCalcMeth=outgoingrayalgorithm.rawApproximation,
                                                               turnOffSolidAngle=self.turnOffSolidAngle)[0]

            self.sinos[element] = sino_value
            self.phantoms[element] = ph_value
        edf_writer = None

        i = 0
        for element in self.elements:
            self.sinos[element] = reconstrutils.decreaseMatSize(self.sinos[element])
            if not edf_writer:
                edf_writer = fabio.edfimage.EdfImage(data=self.sinos[element],
                                                     header={"tata": "toto"})
            else:
                edf_writer.appendFrame(self,
                                       data=self.sinos[element],
                                       header={"tata": "toto"})
            i = i + 1

        edf_writer.write(self.sino_file)

        return {self.sino_file: self.sinos}

    def setUpSelfAbs(self):
        """
        Generate the self absorption matrice (matrice of absorption for the
        outgoing beams)
        """
        self.compositionAndDict = structs.Materials(
            matComposition=structs.MatComposition(
                fileInfo=fileInfo.MatrixFileInfo(
                        file_path=os.path.join(self.tempdir, 'materials.h5'),
                        data_path=structs.MatComposition.MAT_COMP_DATASET),
                data=self.matOfPhysElmts,
            ),
            materials=structs.MaterialsDic(
                fileInfo=fileInfo.DictInfo(
                    file_path=os.path.join(self.tempdir, 'materials.h5'),
                    data_path=structs.MaterialsDic.MATERIALS_DICT),
                data=configdict.ConfigDict(defaultdict=self.materials)
            )
        )
        self.compositionAndDict.save()

    def setUpConfigFile(self, sinoFiles):
        """
        Create the .cfg file in order to use the config interpreter
        """
        self.cfg_fluo_file = os.path.join(self.tempdir, "fluo_cfg.cfg")

        conf = config.FluoConfig()
        # general stuff, absMat and interactionMat
        conf.isAbsMatASinogram = self._buildFromSinogram
        conf.absMat = self.absMat
        conf.materials = self.compositionAndDict
        conf.minAngle = self.minAngle
        conf.maxAngle = self.maxAngle
        conf.projections = str(self.firstProjection) + ':' + str(self.lastProjection)
        conf.includeLastProjection = self.includeLastProjection

        # fluo sinograms path and info
        for filePath in sinoFiles:
            iSino = 0
            sinograms = []
            for element in sinoFiles[filePath]:
                conf.addSinogram(structs.FluoSino(
                    name=element,
                    fileInfo=fileInfo.MatrixFileInfo(
                        file_path=filePath,
                        data_slice=(iSino, )),
                    physElmt=element,
                    ef=self.EF[element],
                    selfAbsMat=None,
                ))
                iSino = iSino + 1

        # FreeART params
        conf.voxelSize = self.voxelSize
        conf.oversampling = self.oversampling
        conf.dampingFactor = self.dampingFactor
        conf.detector = self.det
        conf.E0 = self.E0

        return conf

    def writeConfigFile(self, conf):
        # write the cfg file
        writer = iniConfigIO.IniConfigWriter()
        writer.write(self.cfg_fluo_file, conf)

    def tearDown(self):
        self.tearDownConfigFile()
        self.tearDownAbsMat()
        self.tearDownInteractionMatrice()
        self.tearDownFluorescenceSinograms()

    def tearDownConfigFile(self):
        if os.path.isfile(self.cfg_fluo_file):
            os.unlink(self.cfg_fluo_file)

    def tearDownAbsMat(self):
        if os.path.isfile(self.absorption_matrix_file):
            os.unlink(self.absorption_matrix_file)

    def tearDownInteractionMatrice(self):
        for element in self.elements:
            if os.path.isfile(self.interaction_matrix_file[element]):
                os.unlink(self.interaction_matrix_file[element])

    def tearDownFluorescenceSinograms(self):
        if os.path.isfile(self.sino_file):
            os.unlink(self.sino_file)

    def tearDownMaterialFile(self):
        if os.path.isfile(self.materials_file):
            os.unlink(self.materials_file)


class ComptonExperimentationExample(FluorescenceExperimentationExample):
    """
    A simple compton experimentation example
    """
    def __init__(self, elements, materials, sheppLoganPartID, densities,
                 oversampling, dampingFactor, nbAngles=360,
                 buildFromSinogram=False, E0=1.0e6, matrixWidth=56,
                 minAngle=0.0, maxAngle=2.0*np.pi, firstProjection=0,
                 lastProjection=-1, includeLastProjection=False):

        EF = {}
        for element in elements:
            EF[element] = E0

        FluorescenceExperimentationExample.__init__(
            self,
            elements,
            materials,
            sheppLoganPartID,
            densities,
            oversampling,
            dampingFactor,
            buildFromSinogram=buildFromSinogram,
            E0=E0,
            EF=EF,
            matrixWidth=matrixWidth,
            nbAngles=nbAngles,
            minAngle=minAngle,
            maxAngle=maxAngle,
            firstProjection=firstProjection,
            lastProjection=lastProjection,
            includeLastProjection=includeLastProjection)

    def writeConfigFile(self, conf):
        conf.reconsType = config._ReconsConfig.COMPTON_ID
        FluorescenceExperimentationExample.writeConfigFile(self, conf)


def createConfigTx(tempdir, sinoI0=None):
    txSinoFile = os.path.join(tempdir, 'txSino.edf')
    sinoData = np.arange(100).reshape(10, 10) + 1.0
    sino = structs.TxSinogram(data=sinoData,
                              fileInfo=fileInfo.MatrixFileInfo(file_path=txSinoFile),
                              name='TxSinogram')
    sino.save()
    txConf = config.TxConfig(
        sinograms=(sino,),
        projections='2:6',
        minAngle=0.0,
        maxAngle=1.2,
        dampingFactor=1.,
        I0=sinoI0)

    return txConf


def createConfigFluo(tempdir, addAbsMat=False, addSelfAbsMat=False,
                     selfAbsMat=None, withMaterials=False, fileExtension='.h5',
                     fileExtensionMat='.h5', sinoI0=None):
    NBFluoSino = 5
    outFileFluo = os.path.join(tempdir, 'fluo.h5')

    detector = structs.Detector(x=1, y=2, z=3, width=0.1)
    absMat = None
    if addAbsMat is True:
        fileName = 'AbsmatMatrix.edf'
        sourceFile = os.path.join(tempdir, fileName)
        file_info = fileInfo.MatrixFileInfo(file_path=sourceFile)
        absMat = structs.AbsMatrix(data=np.arange(100).reshape(10, 10) + 1.0,
                                   fileInfo=file_info)
        absMat.save()
    conf = config.FluoConfig(outBeamCalMethod=None,
                             I0=sinoI0,
                             absMat=absMat,
                             isAbsMatASinogram=False,
                             detector=detector,
                             minAngle=0.0,
                             maxAngle=1.2,
                             dampingFactor=1.,
                             centerOfRotation=6.,
                             acquiInverted=True)

    def _addFluoSinograms(fileExtension, addSelfAbsMat=False,
                          selfAbsMat=None):
        def createFileInfo(fileExtension, name, data):
            if fileExtension is None:
                # TODO : try with official one
                # in this case all data are stored in the same .h5 file
                sourceFile = None
                path = "/data/" + name + str(iSino)
                file_info = fileInfo.MatrixFileInfo(file_path=outFileFluo,
                                                    data_path=path)
                sourceFile = outFileFluo
                url = DataUrl(file_path=sourceFile, data_path=path)

                h5utils.createH5WithDataSet(url=url,
                                            data=dataSino,
                                            mode='a')

            elif fileExtension in ('.h5', '.hdf5'):
                # generate a new h5 file
                fileName = name + 'File' + str(iSino) + '.h5'
                sourceFile = os.path.join(tempdir, fileName)

                path = "/data/" + name + str(iSino)
                url = DataUrl(file_path=sourceFile, data_path=path)
                h5utils.createH5WithDataSet(url=url,
                                            data=dataSino)
                file_info = fileInfo.MatrixFileInfo(file_path=sourceFile,
                                                    data_path=path)
            elif fileExtension.lower() == '.edf':
                fileName = name + 'File' + str(iSino) + '.edf'
                sourceFile = os.path.join(tempdir, fileName)

                file_info = fileInfo.MatrixFileInfo(file_path=sourceFile)
                reconstrutils.saveMatrix(data=dataSino, fileName=sourceFile)
            else:
                raise NotImplementedError('Not managed for other types')

            return file_info, sourceFile

        for iSino in range(NBFluoSino):
            dataSino = np.arange(100).reshape(10, 10)
            selfAbsMat = None

            fileInfoSino, sourceFileSino = createFileInfo(fileExtension,
                                                          'fluoSino',
                                                          dataSino)
            if addSelfAbsMat is True:
                dataSelfAbsMat = dataSino
                fileInfoSAM, sourceFileSAM = createFileInfo(fileExtension,
                                                            'selfAbsMat',
                                                            dataSelfAbsMat)
                selfAbsMat = structs.SelfAbsMatrix(fileInfo=fileInfoSAM,
                                                   data=dataSelfAbsMat)
            sinogram = structs.FluoSino(fileInfo=fileInfoSino,
                                        data=dataSino,
                                        name='Sinogram_' + str(iSino),
                                        physElmt='O',
                                        ef=12.0,
                                        selfAbsMat=selfAbsMat)
            conf.addSinogram(sinogram)

    def _addMaterials(fileExtensionMat, filePath):
        composition = np.ndarray((100, 100), dtype='S10')
        composition[:, :] = ''
        composition[20:25, :] = 'mat1'
        composition[:10, 10:23] = 'mat2'
        materials = {
            'mat1':
                OrderedDict({
                    'CompoundList': ['H', 'O'],
                    'CompoundFraction': [0.42, 0.36],
                    'Density': 1.0,
                    'Thickness': 1.0,
                    'Comment': 'wothout'
                }),
            'mat2':
                OrderedDict({
                    'CompoundList': ['Y', 'O'],
                    'CompoundFraction': [0.42, 0.0],
                    'Density': 0.2,
                    'Thickness': 0.5,
                    'Comment': ''
                }),
        }
        if fileExtensionMat is None:
            # save composition
            url = DataUrl(file_path=filePath,
                          data_path=structs.MatComposition.MAT_COMP_DATASET)
            h5utils.createH5WithDataSet(url=url,
                                        data=composition,
                                        mode='a')

            dictdump.dicttoh5(h5file=filePath,
                              treedict=materials,
                              h5path=structs.MaterialsDic.MATERIALS_DICT.lstrip(
                                  "::"),
                              mode='a')

            file_info_dict = fileInfo.MatrixFileInfo(file_path=filePath,
                                                     data_path=structs.MaterialsDic.MATERIALS_DICT)
            file_info_comp = fileInfo.MatrixFileInfo(file_path=filePath,
                                                     data_path=structs.MatComposition.MAT_COMP_DATASET)
        elif fileExtensionMat == '.h5':
            # Then both materials and materials composition are stored in the same .h5
            h5File = os.path.join(tempdir, 'matCompo.h5')
            path = "/data/matCompositions"
            url = DataUrl(file_path=h5File, data_path=path)
            h5utils.createH5WithDataSet(url=url,
                                        data=composition)
            file_info_comp = fileInfo.MatrixFileInfo(file_path=h5File,
                                                     data_path=path)

            h5File = os.path.join(tempdir, 'matDict.h5')
            path = '/materials'
            dictdump.dicttoh5(h5file=h5File,
                              treedict=materials,
                              h5path=path)

            file_info_dict = fileInfo.DictInfo(file_path=h5File, data_path=path)

        elif fileExtensionMat == '.npy':
            npzFile = os.path.join(tempdir, 'matCompo.npy')
            file_info_comp = fileInfo.MatrixFileInfo(npzFile)
            np.save(npzFile, composition)

            dicFile = os.path.join(tempdir, 'matDict.dict.ini')
            tmp = configdict.ConfigDict(materials)
            tmp.write(dicFile)

            assert os.path.isfile(dicFile)
            p = dictdump.load(dicFile)
            assert (p == tmp)
            file_info_dict = fileInfo.MatrixFileInfo(dicFile)
        else:
            raise ValueError('Extansion not managed')

        conf.materials = structs.Materials(
            materials=structs.MaterialsDic(fileInfo=file_info_dict,
                                           data=materials),
            matComposition=structs.MatComposition(fileInfo=file_info_comp,
                                                  data=composition)
        )

    _addFluoSinograms(fileExtension, addSelfAbsMat, selfAbsMat)
    if withMaterials:
        _addMaterials(fileExtensionMat, outFileFluo)

    return conf



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/
__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "01/09/2016"

import unittest
import numpy
from freeart.unitsystem import metricsystem
from freeart.utils import testutils
import silx
if silx._version.MINOR < 7:
    from silx.test.utils import ParametricTestCase
else:
    from silx.utils.testutils import ParametricTestCase

class SheppLoganReconstructionTest(ParametricTestCase):
    """
    This is the unit test class to run a simple projection on the SheppLogan and then make his reconstruction to make sure we find back the same values
    """
    detectorRadius = 10 * metricsystem.cm

    nbIterations = 3
    oversampling = 4

    def test_SheppLogam(self):
        """
        The main function of the SheppLoganReconstruction unit test
        """
        for precision in (numpy.float32, numpy.float64):
            for voxelSize in (2.0*metricsystem.cm, 0.5*metricsystem.mm):
                with self.subTest(precision=precision, voxelSize=voxelSize):
                    rtol = 0.2 if precision is numpy.float64 else 0.3
                    shepLogan = testutils.SheppLoganReconstruction()
                    detPos = (1000 * metricsystem.cm, 0, 0)
                    detSetup = [(detPos, self.detectorRadius)]
                    sino, angles, absMat, selfAbsMat, initialPhantom = shepLogan.produce_SheppLogan_sinogram(32,
                                                                                                             _oversampling=self.oversampling,
                                                                                                             _anglesNb=360,
                                                                                                             _detSetup=detSetup,
                                                                                                             _voxelSize=voxelSize)
                    reconstructedPhantom = shepLogan.make_reconstruction(_sinogram=sino.astype(precision),
                                                                         _angles=angles,
                                                                         _absMat=absMat.astype(precision),
                                                                         _selfAbsMat=selfAbsMat.astype(precision),
                                                                         _detSetup=detSetup,
                                                                         _nbIter=self.nbIterations,
                                                                         _oversampling=self.oversampling,
                                                                         _dampingFactor=0.2,
                                                                         _voxelSize=voxelSize)

                    numpy.testing.assert_allclose(reconstructedPhantom,
                                                  initialPhantom,
                                                  rtol=rtol,
                                                  atol=1e-3)


def suite():
    test_suite = unittest.TestSuite()
    test_suite.addTest(unittest.defaultTestLoader.loadTestsFromTestCase(SheppLoganReconstructionTest))
    return test_suite


if __name__ == '__main__' and __package__ is None:
    unittest.main(defaultTest="suite")



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/
__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "01/10/2016"

import os
import tempfile
import unittest
from freeart.interpreter import GlobalConfigInterpreter
import numpy
from freeart.utils import reconstrutils, testutils, genph
import freeart
from freeart.configuration import config, structs, fileInfo
from freeart.configuration.iniConfigIO import IniConfigWriter
from freeart.unitsystem import metricsystem

import numpy as np
import logging
_logger = logging.getLogger("unitests")

class testTxAngle(unittest.TestCase):
    """Test that we are able to generate a valid transmission reconstruction using a cfg file and the interpreter"""
    oversampling = 1
    voxelSize = 5.0*metricsystem.cm
    phantom_width = 56
    dampingFactor = 1.0/float(phantom_width)
    nbAngles = 360

    nbIteration = 1

    def setUp(self):
        
        # deal with files
        self.tempdir = tempfile.mkdtemp()        
        self.cfg_file_min_max = os.path.join(self.tempdir, "tmp_cfg_file_min_max.cfg")    
        self.cfg_file_first_last = os.path.join(self.tempdir, "tmp_cfg_file_first_last.cfg")    
        self.sino_10_30 = os.path.join(self.tempdir, "sino_10_30.edf")    
        self.sino_tx = os.path.join(self.tempdir, "sino_tx.edf")    
        self.phantom_out_tx = os.path.join(self.tempdir, "phantom_out_tx.edf")    

        # create and write a sino file
        phGenerator = genph.PhantomGenerator()
        self.sheppLogan_phantom_tx = phGenerator.get2DPhantomSheppLogan(self.phantom_width)
        # move it to 3D
        self.sheppLogan_phantom_tx.shape = (self.sheppLogan_phantom_tx.shape[0],
                                            self.sheppLogan_phantom_tx.shape[1],
                                            1)

        # build standard sinogram
        al = freeart.TxFwdProjection(self.sheppLogan_phantom_tx,
                                     minAngle=0,
                                     maxAngle=2.0*numpy.pi,
                                     anglesNb=self.nbAngles)
        al.setOverSampling(self.oversampling)
        al.setVoxelSize(self.voxelSize)
        al.setRandSeedToZero(True)
        sinogram = al.makeSinogram()[0]

        sinogram_10_30 = sinogram[:,10:31,]
        assert(sinogram_10_30.shape[1] == 21)

        reconstrutils.saveSinogram(sinogram_10_30, self.sino_10_30)
        reconstrutils.saveSinogram(sinogram, self.sino_tx)
        assert(os.path.isfile(self.sino_10_30))
        assert(os.path.isfile(self.sino_tx))

        # define parameters of reconstruction
        writer = IniConfigWriter()

        conf = config.TxConfig()
        conf.reconsType = "Transmission"
        conf.sinograms = (
            structs.TxSinogram(
                fileInfo=fileInfo.MatrixFileInfo(file_path=self.sino_10_30)),
        )
        conf.oversampling = self.oversampling
        conf.dampingFactor = self.dampingFactor
        conf.includeLastProjection = True
        conf.minAngle = 10.0 * 2.0 * np.pi / 360.0
        conf.maxAngle = 30.0*2.0*np.pi/360.0
        conf.projections = '0:21'
        conf.voxelSize = self.voxelSize
        writer.write(filePath=self.cfg_file_min_max,
                     reconsConfiguration=conf)
        conf.includeLastProjection = False
        conf.minAngle = 0.0
        conf.maxAngle = 2.0*np.pi
        conf.projections = '10:31'
        conf.sinograms = (
            structs.TxSinogram(
                fileInfo=fileInfo.MatrixFileInfo(file_path=self.sino_tx)),
        )
        writer.write(filePath=self.cfg_file_first_last,
                     reconsConfiguration=conf)

        # build the reconstruction of reference
        angles = np.linspace(10.0*2.0*np.pi/360.0, 30.0*2.0*np.pi/360.0, 21)
        checkAngles = np.linspace(0.0, 2.0*np.pi, self.nbAngles)
        checkAngles = checkAngles[:-1]
        assert(len(angles) == 21)
        # assert(np.array_equal(checkAngles[10:30], angles))
        I0 = 1.0
        alRecons = freeart.TxBckProjection(-np.log(sinogram_10_30/I0), angles)
        alRecons.setRandSeedToZero(True)
        alRecons.setOverSampling(self.oversampling)
        alRecons.setVoxelSize(self.voxelSize)
        alRecons.setDampingFactor(self.dampingFactor)
        # make the reconstruction
        self.refReconstruction = alRecons.iterate(self.nbIteration)

    def tearDown(self):
        # clean cfg
        if os.path.isfile(self.cfg_file_min_max):
            os.unlink(self.cfg_file_min_max)

        if os.path.isfile(self.cfg_file_first_last):
            os.unlink(self.cfg_file_first_last)

        # clean edf
        if os.path.isfile(self.sino_10_30):
            os.unlink(self.sino_10_30)

        if os.path.isfile(self.sino_tx):
            os.unlink(self.sino_tx)            

    def testMinAndMaxAngle(self):
        interpreter = GlobalConfigInterpreter(self.cfg_file_min_max)
        interpreter.setRandSeedToZero(True)
        interpreter.iterate(self.nbIteration)
        k0 = list(interpreter.getReconstructionAlgorithms().keys())[0]
        phantom = interpreter.getReconstructionAlgorithms()[k0].getPhantom()
        
        sino = interpreter.interpreter.getSinograms()[0]
        self.assertTrue(len(interpreter.interpreter.angles) == 21)

        self.assertAlmostEqual(interpreter.interpreter.angles[0], 10.0*2.0*np.pi/360.0)
        self.assertAlmostEqual(interpreter.interpreter.angles[-1], 30.0*2.0*np.pi/360.0)

    def testFirstAndLastProjection(self):
        # pass
        interpreter = GlobalConfigInterpreter(self.cfg_file_first_last)
        interpreter.setRandSeedToZero(True)
        interpreter.iterate(self.nbIteration)
        k0 = list(interpreter.getReconstructionAlgorithms().keys())[0]
        phantom = interpreter.getReconstructionAlgorithms()[k0].getPhantom()

        sino = interpreter.interpreter.getSinograms()[0]
        self.assertTrue(len(interpreter.interpreter.angles) == 21)
        self.assertAlmostEqual(interpreter.interpreter.angles[0], 10.0*2.0*np.pi/360.0)
        self.assertAlmostEqual(interpreter.interpreter.angles[-1], 30.0*2.0*np.pi/360.0)


class TestFluoAngle(unittest.TestCase):
    """Test that we are able to generate a valid fluorescence reconstruction using a cfg file and the interpreter"""

    def setUp(self):
        materialSteel = {'Comment':"No comment",
                 'CompoundList':['Cr', 'Fe', 'Ni'],
                 'CompoundFraction':[18.37, 69.28, 12.35],
                 'Density':1.0,
                 'Thickness':1.0}

        materials1 = {"Steel":materialSteel}
        sheppLoganPartID1 = {"Steel":2}
        densities = {"Steel":1.0}
        
        EF = {"Ni":2.0e6}
        self.elements=['Ni']
        E0 = 1.0e7

        matrixWidth = 56
        dampingFactor = float(matrixWidth) / 1.0
        self.experimentation1 = testutils.FluorescenceExperimentationExample(
            self.elements, materials1, sheppLoganPartID1, densities,
            oversampling=20, dampingFactor=dampingFactor, E0=E0, EF=EF, nbAngles=360,
            minAngle=0.0, maxAngle=2.0*np.pi, firstProjection=10, lastProjection=30, 
            includeLastProjection=False, matrixWidth=matrixWidth)
        self.experimentation1.setUp()
        self.experimentation1.voxelSize = 0.2*metricsystem.mm

        self.experimentation2 = testutils.FluorescenceExperimentationExample(
            self.elements, materials1, sheppLoganPartID1, densities,
            oversampling=20, dampingFactor=dampingFactor, E0=E0, EF=EF,
            nbAngles=21, minAngle=10.0*np.pi*2.0/360.0,
            maxAngle=30.0*np.pi*2.0/360.0, firstProjection=0,
            lastProjection=20, includeLastProjection=True)
        self.experimentation2.setUp()
        self.experimentation2.voxelSize = 0.2 * metricsystem.mm

    def testAngle(self):
        exp1 = self.experimentation1
        interpreter1 = GlobalConfigInterpreter(exp1.cfg_fluo_file)
        interpreter1.setRandSeedToZero(True)
        interpreter1.iterate(1)

        exp2 = self.experimentation2
        interpreter2 = GlobalConfigInterpreter(exp2.cfg_fluo_file)
        interpreter2.setRandSeedToZero(True)
        interpreter2.iterate(1)        
        
        algos1 = interpreter1.getReconstructionAlgorithms()
        algos2 = interpreter2.getReconstructionAlgorithms()

        sinoI1 = interpreter1.interpreter.getSinograms()[0]
        sinoI2 = interpreter2.interpreter.getSinograms()[0]
        self.assertTrue(sinoI1.data.shape == sinoI2.data.shape)
        self.assertTrue(sinoI1.data.shape[1] == 20)
        assert('Ni' in algos1 and 'Ni' in algos2)
        numpy.testing.assert_allclose(algos1['Ni'].getPhantom(), algos2['Ni'].getPhantom(), atol=1e-20 )

    def tearDown(self):
        self.experimentation1.tearDown()
        self.experimentation2.tearDown()
        del self.experimentation1
        del self.experimentation2


def suite():
    test_suite = unittest.TestSuite()
    test_suite.addTest(unittest.defaultTestLoader.loadTestsFromTestCase(testTxAngle))
    test_suite.addTest(unittest.defaultTestLoader.loadTestsFromTestCase(TestFluoAngle))
    
    return test_suite


if __name__ == '__main__':
    unittest.main(defaultTest="suite")



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/

__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "27/10/2017"


import h5py
import os
from silx.io import dictdump
from silx.io.url import DataUrl


def createH5WithDataSet(url, data, mode='w-', overwrite=True):
    """Create a simple file with one data set at the location path"""
    def _getOrCreateNext(topLevelNode, nextGrp):
        if nextGrp in topLevelNode.keys():
            return topLevelNode[nextGrp]
        else:
            return topLevelNode.create_group(nextGrp)

    assert isinstance(url, DataUrl)
    # there is an hack for dict, we have to go through dictdump.dicttoh5
    if isinstance(data, dict):
        dictdump.dicttoh5(h5file=url.file_path(),
                          treedict=data,
                          h5path=url.data_path(),
                          mode='a',
                          overwrite_data=overwrite)
    else:
        file = h5py.File(name=url.file_path(), mode=mode)
        grps = url.data_path().lstrip('/').split('/')
        datasetname = grps[-1]
        grps=grps[:-1]
        node = file
        for grp in grps:
            node = _getOrCreateNext(node, grp)

        if not datasetname in node:
            node.create_dataset(name=datasetname, data=data)
        elif overwrite is True:
            del node[datasetname]
            node[datasetname] = data
        else:
            raise IOError('Unable to create link (Name already exists)')
        file.close()
        assert (os.path.isfile(url.file_path()))



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import sys
from freeart.interpreter import GlobalConfigInterpreter
from freeart.utils.reconstrutils import decreaseMatSize
from silx.gui.plot import Plot2D
from silx.gui import qt
import argparse
import logging

_logger = logging.getLogger(__name__)

def getinputinfo():
    return "freeart interpreter [projectfile.cfg] [nbIteration]"


def main(argv):
    parser = argparse.ArgumentParser(description=__doc__)
    parser.add_argument(
        'configuration_file',
        help='File containing the configuration of the reconstruction')
    parser.add_argument(
        'nb_iteration',
        help='number of iteration to execute on the reconstruction')
    parser.add_argument(
        '--debug',
        dest="debug",
        action="store_true",
        default=False,
        help='Set logging system in debug mode')

    options = parser.parse_args(argv[1:])

    cfgFile = options.configuration_file
    nbIteration = options.nb_iteration

    _logger.info('create the freeart interpreter to run reconstruction')
    interpreter = GlobalConfigInterpreter(filePath=cfgFile)
    _logger.info('start iterations')
    interpreter.iterate(nbIteration)
    _logger.info('iterations finished')

    global app  # QApplication must be global to avoid seg fault on quit
    app = qt.QApplication([])

    for algo in interpreter.getReconstructionAlgorithms():
        reconsPh = interpreter.getReconstructionAlgorithms()[algo].getPhantom()
        plot = Plot2D(parent=None)
        plot.addImage(decreaseMatSize(reconsPh))
        plot.show()

    app.exec_()
    exit(0)

if __name__ == "__main__":
    main(sys.argv)



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/
__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "16/11/2017"

import unittest
import numpy as np
import silx
if silx._version.MINOR < 7:
    from silx.test.utils import ParametricTestCase
else:
    from silx.utils.testutils import ParametricTestCase
import freeart
from freeart.unitsystem import metricsystem
from freeart.utils import raypointsmethod, testutils
import numpy
import logging

_logger = logging.getLogger("unitests")


class TestTxProjection_2D(ParametricTestCase):
    """
    the unit test related to the transmission projector in a classical 2D case
    """

    oversampling = 1
    phDimX = 101
    phDimY = 101
    phDimZ = 1

    def test_projector_1(self):
        """
        A simple test to try the projector on a single attenuating voxel
        """
        xpos = 50
        ypos = 50
        zpos = 0

        for precision in [numpy.float32, numpy.float64]:
            for I0 in [2.0, 3.5e3, 9.8e8]:
                for voxelSize in [1.0*metricsystem.centimeter, 2.3*metricsystem.millimeter, 0.56*metricsystem.meter]:
                    # all the absorption factor to test
                    for absorption_factor in np.linspace(0.2, 10, 5):
                        with self.subTest(precision=precision, I0=I0, voxelSize=voxelSize, absorptionFactor=absorption_factor):
                            # create the simple phantom
                            phantom = np.zeros((self.phDimX, self.phDimY, self.phDimZ),
                                               dtype=precision)
                            phantom[ypos, xpos, zpos] = absorption_factor  # unit : g/cm2

                            # create the sinogram
                            al = freeart.TxFwdProjection(phantom,
                                                         minAngle=0,
                                                         maxAngle=2.0*np.pi,
                                                         anglesNb=9)
                            al.setOverSampling(8)
                            al.setVoxelSize(voxelSize)
                            al.setI0(I0)
                            sinogram, angles = al.makeSinogram()

                            # check that the max value <=> absorption_factor on the voxel == theoretical value
                            dist_in_mat = 1.0 * voxelSize    # distance : 1.0 cm
                            theoreticalValue = np.exp(-dist_in_mat*absorption_factor)*I0
                            # we are only interested of the sinogram obtained for the central ray for the first angle ( 0 degree ).
                            freeARTValue = sinogram[0][0][xpos]

                            if precision == numpy.float64:
                                self.assertAlmostEqual(first=precision(theoreticalValue),
                                                       second=freeARTValue)

    def test_projector_3(self):
        """
        a simple test to check values for a uniform filled in a 3x3 voxels phantom
        """
        voxelSize = 1.0 * metricsystem.centimeter
        for precision in [numpy.float32, numpy.float64]:
            for absorption_factor in np.linspace(0.2, 10, 15):
                with self.subTest(precision=precision, absorption_factor=absorption_factor):

                    # create the simple phantom
                    phantom = np.zeros((self.phDimX, self.phDimY, self.phDimZ), dtype=precision)
                    phantom[49:52, 49:52, 0] = absorption_factor  # unit : g/cm2

                    # create the sinogram
                    al = freeart.TxFwdProjection(phantom,
                                                 minAngle=0,
                                                 maxAngle=2.0*np.pi,
                                                 anglesNb=9)
                    al.setOverSampling(self.oversampling)
                    al.setRandSeedToZero(True)
                    al.setVoxelSize(voxelSize)

                    sinogram, angles = al.makeSinogram()

                    # check that the max value <=> absorption_factor on the voxel == theoretical value
                    dist_in_mat = 3.0 * voxelSize
                    theoreticalValue = np.exp(-dist_in_mat*absorption_factor)
                    # we are only interested of the sinogram obtained for the central ray for the first angle ( 0 degree ).

                    # we are only interested of the sinogram obtained for the central ray for the first angle ( 0 degree ).
                    freeARTValue1 = sinogram[0][0][49]
                    freeARTValue2 = sinogram[0][0][50]
                    freeARTValue3 = sinogram[0][0][51]

                    # when the incoming beam hit the sample at 45 degree
                    sinoVal45_degree = sinogram[0][1][50]
                    freeARTValue_45 = sinoVal45_degree
                    # when the incoming beam hit the sample at 90 degree
                    sinoVal90_degree = sinogram[0][2][50]
                    freeARTValue_90 = sinoVal90_degree

                    # check that the max value <=> absorptionFactor on the voxel == theoretical value
                    theoreticalValue = np.exp(-dist_in_mat*absorption_factor)

                    theoreticalValue_90 = np.exp(-dist_in_mat*absorption_factor)*voxelSize
                    theoreticalValue_45 = np.exp(-np.sqrt(dist_in_mat*dist_in_mat + dist_in_mat*dist_in_mat)*absorption_factor)*voxelSize

                    self.assertEqual(freeARTValue1, freeARTValue2)
                    self.assertEqual(freeARTValue1, freeARTValue3)
                    self.assertAlmostEqual(theoreticalValue, freeARTValue1)

                    self.assertAlmostEqual(theoreticalValue_90, freeARTValue_90)
                    self.assertAlmostEqual(theoreticalValue_45, freeARTValue_45)

    def test_projector_1_shifted(self):
        """
        Test one off-center voxel if we find back the theorical value
        """
        xpos = 26
        ypos = 65
        voxelSize = 1.0 * metricsystem.centimeter

        for absorptionFactor in np.linspace(0.5, 10, 10):
            # create the simple phantom
            phantom = np.zeros((self.phDimX, self.phDimY, self.phDimZ),
                               dtype=np.float64)
            phantom[ypos, xpos, 0] = absorptionFactor  # unit : g/cm2

            # create the sinogram
            al = freeart.TxFwdProjection(phantom,
                                         minAngle=0,
                                         maxAngle=2.0*np.pi,
                                         anglesNb=9)
            al.setRandSeedToZero(True)
            al.setVoxelSize(voxelSize)
            sinogram, angles = al.makeSinogram()

            # check that the max value <=> absorptionFactor on the voxel == theoretical value
            distInMat = 1.0 * voxelSize
            theoreticalValue = np.exp(- distInMat*absorptionFactor)
            # we are only interested of the sinogram obtained for the central ray for the first angle ( 0 degree ).
            freeARTValue = sinogram[0][0][xpos]

            _logger.info("theoreticalValue = " + str(theoreticalValue))
            _logger.info("freeARTValue = " + str(freeARTValue))

            self.assertAlmostEqual(theoreticalValue, freeARTValue)


class TestTxRecontruction_2D(ParametricTestCase):
    """Test that the reconstruction of the transmission work"""
    phDimX = 100
    phDimY = 100
    phDimZ = 1

    def test_reconstruction_lines_y(self):
        """
        Test the reconstruction of a phantom containig a line of voxels set
        with a uniform absorption factor
        """
        absToTest = [10.0]

        oversampling = 8
        nbIterations = 2
        I0 = 2.3e3
        voxelSize = 2.0 * metricsystem.mm

        for precision in [numpy.float32, numpy.float64]:
            for absorptionFactor in absToTest:
                with self.subTest(precision=precision, absorption_factor=absorptionFactor):

                    # make the sinogram
                    initialPhantom = np.zeros((self.phDimX, self.phDimY, self.phDimZ),
                                              dtype=precision)
                    initialPhantom[20:64, 15, 0] = absorptionFactor

                    methodGBeams = raypointsmethod.withInterpolation

                    # create the sinogram
                    alConstr = freeart.TxFwdProjection(initialPhantom,
                                                       minAngle=0,
                                                       maxAngle=2.0*np.pi,
                                                       anglesNb=360)
                    alConstr.setRandSeedToZero(True)
                    alConstr.setOverSampling(oversampling)
                    alConstr.setRayPointCalculationMethod(methodGBeams)
                    alConstr.setI0(I0)
                    alConstr.setVoxelSize(voxelSize)
                    sinogram, angles = alConstr.makeSinogram()

                    # make the reconstruction
                    alRecons = freeart.TxBckProjection(-np.log(sinogram/I0), angles)
                    alRecons.setRandSeedToZero(True)
                    alRecons.setOverSampling(oversampling)
                    alRecons.setRayPointCalculationMethod(methodGBeams)
                    alRecons.setI0(I0)
                    alRecons.setVoxelSize(voxelSize)
                    alRecons.setDampingFactor(0.2)
                    # make the reconstruction
                    reconstructedPhantom = alRecons.iterate(nbIterations)

                    _logger.info("absorptionFactor = " + str(absorptionFactor))
                    _logger.info("max value in the reconstructed = " + str(np.amax(reconstructedPhantom)))

                    if precision is numpy.float64:
                        np.testing.assert_allclose(reconstructedPhantom,
                                                   initialPhantom,
                                                   rtol=0.1,
                                                   atol=0.01)

    def test_reconstruction_symetric(self):
        """
        Test the reconstruction process on a phantom containing
        two elements.
        The background absorption is 0.
        Then on top of this background we have a cross with a value of
        absorption set to crossAbsorption.
        And on the higher level we have a disc with a value of absorption set
        to discAbsorption.
        """
        oversampling = 5
        nbIterations = 4

        methodGBeams = raypointsmethod.withInterpolation

        discRadius = 20
        I0 = 2.3e5
        voxelSize = 0.2 * metricsystem.millimeter

        # create the mask for the phantom
        crossMask = np.zeros((self.phDimX, self.phDimY), dtype=np.float64)
        crossMask[40:60, :] = 1.0
        crossMask[:, 40:60] = 1.0

        discMask = np.zeros((self.phDimX, self.phDimY), dtype=np.float64)
        discOrigin = (self.phDimX/2.0, self.phDimY/2.0)
        for x in np.arange(0, discMask.shape[0]):
            for y in np.arange(discMask.shape[1]):
                if testutils.getDistance_2D((x, y), discOrigin) <= discRadius:
                    discMask[x, y] = 1.0

        for precision in (numpy.float32, numpy.float64):
            # test for different absorption level on the disc
            for discAbsorption in (4.e-2, 5.2):
                # test for different absorption level on the cross
                for crossAbsorption in (1.e-2, 2.2):
                    with self.subTest(precision=precision,
                                      discAbsorption=discAbsorption,
                                      crossAbsorption=crossAbsorption):

                        # update the phantom
                        initialPhantom = np.zeros((self.phDimX, self.phDimY, self.phDimZ),
                                                  dtype=np.float64)
                        initialPhantom[crossMask == 1] = crossAbsorption
                        initialPhantom[discMask == 1] = discAbsorption

                        # remove the blind voxel due to the fact that freeART is reducing the projection area
                        # to the higher circle contained in the phantom (simplification)
                        initialPhantom = testutils.removeAvoidVoxel(initialPhantom)

                        alProjo = freeart.TxFwdProjection(initialPhantom,
                                                          minAngle=0,
                                                          maxAngle=2.0*np.pi,
                                                          anglesNb=360)
                        alProjo.setOverSampling(oversampling)
                        alProjo.setRayPointCalculationMethod(methodGBeams)
                        alProjo.setVoxelSize(voxelSize)
                        alProjo.setI0(I0)
                        sinogram, angles = alProjo.makeSinogram()

                        alRecons = freeart.TxBckProjection(-np.log(sinogram/I0),
                                                           angles)
                        alRecons.setOverSampling(oversampling)
                        alRecons.setRayPointCalculationMethod(methodGBeams)
                        alRecons.setDampingFactor(0.05)
                        alRecons.setRandSeedToZero(True)
                        alRecons.setVoxelSize(voxelSize)
                        alRecons.setI0(I0)
                        reconstructedPhantom = alRecons.iterate(nbIterations)

                        # filtering external voxel for the tests
                        reconstructedPhantom[0, :] = 0
                        reconstructedPhantom[:, 0] = 0
                        reconstructedPhantom[:, self.phDimX-1] = 0
                        reconstructedPhantom[self.phDimY-1, :] = 0

                        initialPhantom[0, :] = 0
                        initialPhantom[:, 0] = 0
                        initialPhantom[:, self.phDimX-1] = 0
                        initialPhantom[self.phDimY-1, :] = 0

                        absoluteTol = 0.2 * (max(crossAbsorption, discAbsorption))

                        _logger.info("test for crossAbsorption = " + str(crossAbsorption) + ", discAbsorption = " + str(discAbsorption))
                        np.testing.assert_allclose(reconstructedPhantom, initialPhantom, rtol=0.1, atol=absoluteTol)


def suite():
    test_suite = unittest.TestSuite()
    test_suite.addTest(unittest.defaultTestLoader.loadTestsFromTestCase(TestTxProjection_2D))
    test_suite.addTest(unittest.defaultTestLoader.loadTestsFromTestCase(TestTxRecontruction_2D))
    return test_suite


if __name__ == '__main__' and __package__ is None:
    unittest.main(defaultTest="suite")



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/
"""Test two steps.
The first one is the interpretation of a cfg file.
The second one is the reconstruction using the freeart core (ART algorithm)
and the python layer dealing with interaction probability and sample materials.
"""


__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "01/09/2016"

import logging
import os
import tempfile
import unittest
import shutil
import numpy
import numpy as np
import freeart
from freeart.configuration.iniConfigIO import IniConfigWriter
from freeart.configuration import config, structs, fileInfo
from freeart.interpreter import GlobalConfigInterpreter
from freeart.unitsystem import metricsystem
from freeart.utils import reconstrutils, testutils, genph
import silx
if silx._version.MINOR < 7:
    from silx.test.utils import ParametricTestCase
else:
    from silx.utils.testutils import ParametricTestCase
import copy

_logger = logging.getLogger("unitests")


class TestConfigInterpreterTx(unittest.TestCase):
    """
    Test that we are able to generate a valid transmission reconstruction using
    a cfg file and the interpreter
    """
    oversampling = 6
    dampingFactor = 0.04
    voxelSize = 2.5*metricsystem.cm
    phantom_width = 56
    nbAngles = 360

    def setUp(self):
        # deal with files
        self.tempdir = tempfile.mkdtemp()
        self.cfg_file_tx = os.path.join(self.tempdir, "tmp_cfg_file_tx.cfg")
        self.cfg_file_tx_2 = os.path.join(self.tempdir, "tmp_cfg_file_tx_2.cfg")
        self.sino_tx = os.path.join(self.tempdir, "sino_tx.edf")
        self.phantom_out_tx = os.path.join(self.tempdir, "phantom_out_tx.edf")

        # create and write a sino file
        phGenerator = genph.PhantomGenerator()
        self.sheppLogan_phantom_tx = phGenerator.get2DPhantomSheppLogan(self.phantom_width)
        # move it to 3D
        self.sheppLogan_phantom_tx.shape = (self.sheppLogan_phantom_tx.shape[0],
                                            self.sheppLogan_phantom_tx.shape[1],
                                            1)

        al = freeart.TxFwdProjection(self.sheppLogan_phantom_tx,
                                     minAngle=0,
                                     maxAngle=2.0*numpy.pi,
                                     anglesNb=self.nbAngles)
        al.setOverSampling(self.oversampling)
        al.setVoxelSize(self.voxelSize)
        al.setRandSeedToZero(True)
        sinogram = al.makeSinogram()[0]

        reconstrutils.saveSinogram(sinogram, self.sino_tx)
        assert(os.path.isfile(self.sino_tx))

        writer = IniConfigWriter()
        # define parameters of reconstruction
        conf = config.TxConfig()
        conf.reconsType = "Transmission"
        conf.sinograms = (
            structs.TxSinogram(
                fileInfo=fileInfo.MatrixFileInfo(self.sino_tx)),
        )
        conf.voxelSize = self.voxelSize
        conf.oversampling = self.oversampling
        conf.dampingFactor = self.dampingFactor
        conf.minAngle = 0.0
        conf.maxAngle = 2.0*np.pi
        conf.projections = str(0) + ":" + str(self.nbAngles-1)
        writer.write(self.cfg_file_tx, conf)

        conf.definitionReduction = 2
        writer.write(self.cfg_file_tx_2, conf)

    def tearDown(self):
        # tx files
        if os.path.isfile(self.cfg_file_tx):
            os.unlink(self.cfg_file_tx)

        if os.path.isfile(self.sino_tx):
            os.unlink(self.sino_tx)

    def testTxFromConfig(self):
        # pass
        interpreter = GlobalConfigInterpreter(self.cfg_file_tx)
        # for unit tests, reset the random generator
        interpreter.setRandSeedToZero(True)
        interpreter.iterate(2)
        k0 = list(interpreter.getReconstructionAlgorithms().keys())[0]
        phantom = interpreter.getReconstructionAlgorithms()[k0].getPhantom()
        numpy.testing.assert_allclose(phantom,
                                      self.sheppLogan_phantom_tx,
                                      rtol=0.5,
                                      atol=3e-3)

    def testIteration(self):
        """
        Make sure that when we call iterate start from the current state
        """
        interpreter = GlobalConfigInterpreter(self.cfg_file_tx)
        # for unit tests, reset the random generator
        interpreter.setRandSeedToZero(True)
        interpreter.iterate(1)
        k0 = list(interpreter.getReconstructionAlgorithms().keys())[0]
        phantomIter1 = interpreter.getReconstructionAlgorithms()[k0].getPhantom().copy()

        interpreter.iterate(1)
        phantomIter2 = interpreter.getReconstructionAlgorithms()[k0].getPhantom()

        self.assertFalse(numpy.allclose(phantomIter1, phantomIter2))

    def testReductionData(self):
        """
        Test that the request for reduction data will be well take into account
        """
        interpreter = GlobalConfigInterpreter(self.cfg_file_tx_2)
        # for unit tests, reset the random generator
        interpreter.setRandSeedToZero(True)
        interpreter.iterate(2)
        k0 = list(interpreter.getReconstructionAlgorithms().keys())[0]
        phantom = interpreter.getReconstructionAlgorithms()[k0].getPhantom()
        self.assertTrue(phantom.shape[1] == self.sheppLogan_phantom_tx.shape[1] / 2)


class TestConfigInterpreterCompton(unittest.TestCase):
    """
    Test that we are able to generate a valid fluorescence reconstruction
    using a cfg file and the interpreter
    """
    materialName = "Steel"
    material = {
        'Comment': "No comment",
        'CompoundList': ['Cr', 'Fe', 'Ni'],
        'CompoundFraction': [18.37, 69.28, 12.35],
        'Density': 1.0,
        'Thickness': 1.0
    }

    materials = {materialName: material}
    sheppLoganPartID = {materialName: 2}
    densities = {materialName: 1.0}

    elements = ['K']

    experimentation = testutils.ComptonExperimentationExample(elements,
                                                              materials,
                                                              sheppLoganPartID,
                                                              densities,
                                                              oversampling=20,
                                                              dampingFactor=0.02,
                                                              nbAngles=360)

    def setUp(self):
        self.experimentation.setUp()

    def tearDown(self):
        self.experimentation.tearDown()

    def testFluoFromConfig(self):
        interpreter = GlobalConfigInterpreter(self.experimentation.cfg_fluo_file)
        # for unit tests, reset the random generator
        interpreter.setRandSeedToZero(True)
        interpreter.iterate(5)

        algos = interpreter.getReconstructionAlgorithms()
        # check parameters from the first algorithm
        for material in algos:
            reconstructed_phantom = algos[material].getPhantom()
            assert(material in self.experimentation.phantoms)
            numpy.testing.assert_allclose(reconstructed_phantom,
                                          self.experimentation.phantoms[material],
                                          rtol=0.5,
                                          atol=5e-2)


class TestConfigInterpreterReconstructionFluo(unittest.TestCase):
    """
    Test that the reconstructions for fluorescence
    (densities*interactionMatrix) are correct
    """

    def setUp(self):
        materialSteel = {
            'Comment': "No comment",
            'CompoundList': ['Cr', 'Fe', 'Ni'],
            'CompoundFraction': [18.37, 69.28, 12.35],
            'Density': 1.0,
            'Thickness': 1.0
        }

        materialTeflon = {
            'Comment': "No comment",
            'CompoundList': ['C1', 'F1'],
            'CompoundFraction': [0.240183, 0.759817],
            'Density': 1.0,
            'Thickness': 1.0
        }

        self.elements = ['Ni', 'K']

        materials = {
            "Steel": materialSteel,
            "Teflon": materialTeflon
        }
        sheppLoganPartID = {
            "Steel": 3,
            "Teflon": 2
        }
        densities = {
            "Steel": 0.5,
            "Teflon": 0.2
        }
        EF = {
            "Ni": 1.0e7,
            "K": 2.0e7
        }
        self.experimentationTwoMat = testutils.FluorescenceExperimentationExample(
            self.elements,
            materials,
            sheppLoganPartID,
            densities,
            oversampling=20,
            dampingFactor=0.02,
            E0=1.0e6,
            EF=EF,
            nbAngles=400)
        self.experimentationTwoMat.setUp()
        self.experimentationTwoMat.voxelSize = 3.0*metricsystem.cm

    def testFluoFromConfigTwoMats(self):
        exp = self.experimentationTwoMat
        interpreter = GlobalConfigInterpreter(exp.cfg_fluo_file)
        # for unit tests, reset the random generator
        interpreter.setRandSeedToZero(True)
        interpreter.iterate(3)

        algos = interpreter.getReconstructionAlgorithms()
        # check parameters from the first algorithm
        for material in algos:
            reconstructed_phantom = algos[material].getPhantom()
            assert(material in exp.phantoms)
            numpy.testing.assert_allclose(reconstructed_phantom,
                                          exp.phantoms[material],
                                          rtol=0.3,
                                          atol=5e-2)

    def tearDown(self):
        self.experimentationTwoMat.tearDown()
        del self.experimentationTwoMat


class TestConfigInterpreterDensitiesFluo(unittest.TestCase):
    """
    Test that the materials densities reconstructed from the fluorescence
    SART are correct.
    """

    def _generateMaterials(self):
        materialSteel = {
            'Comment': "No comment",
            'CompoundList': ['Cr', 'Fe', 'Ni'],
            'CompoundFraction': [18.37, 69.28, 12.35],
            'Density': 1.0,
            'Thickness': 1.0
        }

        self.sheppLogan = {"Steel": 2}
        """this sheppLogan will be sued to generate concentrations"""

        import fisx
        from fisx import Elements
        elements = Elements()

        matSteel = fisx.Material('Steel',
                                 materialSteel['Density'],
                                 materialSteel['Thickness'])
        matSteel.setComposition(dict(zip(materialSteel['CompoundList'],
                                         materialSteel['CompoundFraction'])))
        elements.addMaterial(matSteel)

        self.materials = {"Steel": materialSteel}
        self.densities = {"Steel": 1.0}

        self.elements = ['K']

    def setUp(self):
        self.E0 = 1.0e6
        self._generateMaterials()

        EF2 = {"K": 1.0e7}
        self.experimentationTwoMat = testutils.FluorescenceExperimentationExample(
                                        elements=self.elements,
                                        materials=self.materials,
                                        sheppLoganPartIDs=self.sheppLogan,
                                        densities=self.densities,
                                        oversampling=20,
                                        dampingFactor=0.03,
                                        E0=self.E0,
                                        EF=EF2,
                                        nbAngles=400)
        self.experimentationTwoMat.setUp()
        self.experimentationTwoMat.voxelSize = 3.0*metricsystem.cm

    def testFluoFromConfigTwoMats(self):
        exp = self.experimentationTwoMat
        interpreter = GlobalConfigInterpreter(exp.cfg_fluo_file)
        # for unit tests, reset the random generator
        interpreter.setRandSeedToZero(True)
        interpreter.iterate(3)
        reconstructions = interpreter.interpreter.getDensityPhantoms()

        # check parameters from the first algorithm
        for element in reconstructions:
            reconstructed_phantom = reconstructions[element]
            assert(element in exp.phantoms)
            numpy.testing.assert_allclose(
                reconstructed_phantom, exp.densityMat, rtol=0.25, atol=5e-2)

    def tearDown(self):
        self.experimentationTwoMat.tearDown()
        del self.experimentationTwoMat


class TestXYReduction(ParametricTestCase):
    """
    Test that the configinterpreter is correctly dealing with the
    reduction of the sinogram
    """

    def setUp(self):
        ParametricTestCase.setUp(self)
        self.sinoI0 = structs.I0Sinogram(data=np.arange(100).reshape(10, 10) + 1.0)
        self.tempdir = tempfile.mkdtemp()
        self.confTx = testutils.createConfigTx(self.tempdir,
                                               sinoI0=copy.copy(self.sinoI0))
        self.txSinograms = self.confTx.sinograms

        self.confFluo = testutils.createConfigFluo(self.tempdir,
                                                   sinoI0=copy.copy(self.sinoI0),
                                                   addSelfAbsMat=True,
                                                   addAbsMat=True)
        self.confFluo.minAngle = 0.2
        self.confFluo.maxAngle = 0.8
        self.fluoSinograms = self.confFluo.sinograms

    def tearDown(self):
        shutil.rmtree(self.tempdir)
        ParametricTestCase.tearDown(self)

    def testTx(self):
        for center in (2, 5, 7):
            for selection in (':', '1:6', '0:2; 4:6'):
                for reductionX in (1, 2, 3, 6):
                    for reductionY in (1, 2, 3, 6):
                        with self.subTest(selection=selection,
                                          reductionDefinition=reductionX,
                                          reductionAngle=reductionY,
                                          center=center):
                            txConf = config.TxConfig(
                                sinograms=copy.deepcopy(self.txSinograms),
                                projections='2:6',
                                minAngle=0.0,
                                maxAngle=1.2,
                                dampingFactor=1.,
                                I0=copy.deepcopy(self.sinoI0))

                            txConf.projections = selection
                            txConf.definitionReduction = reductionX
                            txConf.projectionNumberReduction = reductionY
                            txConf.center = center
                            gci = GlobalConfigInterpreter(filePath=None,
                                                          config=txConf)
                            gci.iterate(1)

    def testFluo(self):
        for center in (2, 5, 7):
            for selection in (':', '1:6', '0:2; 4:6'):
                for reductionX in (1, 2, 3, 6):
                    w = center if center < 5 else 10 - center
                    currentWidth = w * 2
                    mod = currentWidth % reductionX
                    mod = 1 if mod > 0 else 0
                    matWidth = int(currentWidth / reductionX) + mod
                    absMatData = numpy.arange(matWidth*matWidth).reshape(matWidth,
                                                                         matWidth)
                    absMat = structs.AbsMatrix(data=copy.copy(absMatData))
                    selfAbsMat = structs.SelfAbsMatrix(data=copy.copy(absMatData))
                    for reductionY in (1, 2, 3, 6):
                        with self.subTest(selection=selection,
                                          reductionDefinition=reductionX,
                                          reductionAngle=reductionY,
                                          center=center):
                            fluoConf = self.confFluo
                            fluoConf.setI0(copy.deepcopy(self.sinoI0))
                            fluoConf.sinograms = copy.deepcopy(self.fluoSinograms)
                            fluoConf.absMat = copy.deepcopy(absMat)
                            assert fluoConf.absMat.data.shape == (matWidth, matWidth)
                            for sino in fluoConf.sinograms:
                                sino.selfAbsMat = copy.deepcopy(selfAbsMat)
                                assert sino.selfAbsMat.data.shape == (matWidth, matWidth)
                            fluoConf.projections = selection
                            fluoConf.definitionReduction = reductionX
                            fluoConf.projectionNumberReduction = reductionY
                            fluoConf.center = center
                            # absorption and self absorption matrices have to
                            # be of the correct shape
                            gci = GlobalConfigInterpreter(filePath=None,
                                                          config=fluoConf)
                            gci.iterate(1)


def suite():
    test_suite = unittest.TestSuite()
    for test in (TestConfigInterpreterTx, TestConfigInterpreterCompton,
                 TestConfigInterpreterCompton,
                 TestConfigInterpreterReconstructionFluo,
                 TestConfigInterpreterDensitiesFluo,
                 TestXYReduction):
        test_suite.addTest(unittest.defaultTestLoader.loadTestsFromTestCase(test))

    return test_suite


if __name__ == '__main__':
    unittest.main(defaultTest="suite")
